from ..catalog import PhaseCenterCatalog
from .. import util
from .. import key

import schedlib as s

def pcread(input_iterator):
    global record_defaults
    record_defaults = {
        "name":    ["",   util.noop],
        "sources": [[],   util.foreach(util.upper)],
        "endcent": [None, util.noop],
    }
    input_iterator.set_defaults(record_defaults, {})

    attribute_to_key = {
        "ctrname": "name",
        "ctrsrcn": "sources"
    }
    
    if s.schcon.debug:
        s.wlog(0, "PCREAD:  Starting to read groups of phase centers")
    
    catalog = PhaseCenterCatalog()

    start = int(s.schsou.ncent)
    index = start
    for record in input_iterator:
        values, present = util.merge_record_with_defaults(
            record, record_defaults, {})
        if "endcent" in present:
            break
        entry = catalog.entries[index]
        entry.set_keyin_values(values, attribute_to_key)
        index += 1

    s.schsou.ncent = index
    catalog.write(range(start, index))



from ..catalog import SatelliteCatalog
from . import parameter
from .. import util

import schedlib as s

record_defaults = {
    "satfile": ["NONE", util.expand_file_name],
    "tlefile": ["NONE", util.expand_file_name],
    "endsat":  [None, util.noop],
}
state_defaults = {
    "satname": ["", util.upper],
    "satnum": [parameter.unset, util.noop],
    "kerfile": ["NONE", util.expand_file_name],
}

def satini(input_iterator):
    global record_defaults, state_defaults
    input_iterator.set_defaults(record_defaults, state_defaults)
    
    if s.schcon.debug:
        s.wlog(0, "SATINI: starting.")
    
    catalog = SatelliteCatalog()
    attribute_to_key = dict(zip(catalog.attributes, catalog.attributes))

    start = int(s.schsat.nsat)
    index = start
    for record in input_iterator:
        values, present = util.merge_record_with_defaults(
            record, record_defaults, state_defaults)
        if "endsat" in present:
            break
        entry = catalog.entries[index]
        entry.set_keyin_values(values, attribute_to_key)
        index += 1
        s.wlog(0, "SATINI: Satellite {} {} {}".format(
            index, entry.satname, entry.satnum))
        for attribute in ["kerfile", "satfile", "tlefile"]:
            s.wlog(0, "'        {}: {}".format(attribute.upper(), 
                                               getattr(entry, attribute)))

        error = False
        if ("satfile" in present) and ("tlefile" in present):
            s.wlog(1, "A satellite can only have a SATFILE or a TLEFILE, "
                   "not both.")
            error = True

        if ("satfile" not in present) and ("tlefile" not in present):
            s.wlog(1, "'SATINI:  A satellite must have a SATFILE or a TLEFILE.")
            error = True

        if error:
            s.wlog(1, "  The problem occurred with satellite {} {}".format(
                index, entry.satname))
            s.wlog(1, "'     SATFILE:   {}".format(entry.satfile))
            s.wlog(1, "'     TLEFILE:   {}".format(entry.tlefile))
            s.errlog(" Exactly one of the file names should be specified.")

    s.schsat.nsat = index
    catalog.write(range(start, index))



import math

iusch = 8
iusrc = 9
iusta = 10
iupk = 25
ilog = 29
unset = -9999.
onesec = 1./86400.
secpday = 86400
raddeg = math.pi / 180.
radds = math.pi / 180. / 3600.
radhs = math.pi / 12. / 3600.
radhr = math.pi / 12

max_seg = 100 # MSEG from sched.inc
sidereal = 1.00273791



from ..catalog import StationCatalog
from . import parameter
from .. import util, key

import schedlib as s

import collections
import itertools
import math

def stread(input_iterator, stdin, mjd1):
    
    if input_iterator.input_ is stdin:
        infile = "Program_input"
        locfile = "NOLOC"
    else:
        infile = input_iterator.input_.name
        locfile = util.f2str(s.schcst.locafile)

    # 100 - 25: magic number copied from stread.f
    s.wlog(0, "STREAD:  Reading station catalog: {}".format(infile[-(100-25):]))
    s.wlog(0, "STREAD:  Reading locations file:  {}".format(locfile))

    if locfile.upper not in ("NONE", "NOLOC"):
        try:
            f = open(locfile, "r")
        except Exception as e:
            noloc = True
        else:
            with f:
                location_keyin_data = key.read_keyfile(f)
            noloc = False
    else:
        noloc = True

    mdb = s.rdcatn.dbx.shape[0]
    if (not noloc) and (len(location_keyin_data) > mdb):
        s.error("RDLOC: Too many stations in location file. Max: {}".format(
            mdb))

    station_state_defaults = {
        "version": ["", util.noop],
    }
    station_record_defaults = {
        "station":  ["NONAME",     util.upper],
        "stcode":   ["XX",         util.noop],
        "dbname":   ["",           util.noop],
        "dbcode":   ["",           util.noop],
        "frame":    ["XX",         util.noop],
        "elev":     [0.,           util.noop],
        "lat":      [0.,           util.multiply_by(parameter.raddeg / 3600.)],
        "long":     [0.,           util.multiply_by(parameter.raddeg /3600.)],
        "zalim":    [None,         util.noop],
        "x":        [0.,           util.noop],
        "y":        [0.,           util.noop],
        "z":        [0.,           util.noop],
        "dxdt":     [0.,           util.noop],
        "dydt":     [0.,           util.noop],
        "dzdt":     [0.,           util.noop],
        "epoch":    [0.,           util.noop],
        "descrip":  ["",           util.noop],
        "control":  ["NONE",       util.upper],
        "dar":      ["NONE",       util.noop],
        "recorder": ["NONE",       util.noop],
        "ndrives":  [2.,           util.noop],
        "nheads":   [1.,           util.noop],
        "disk":     ["",           util.noop],
        # two spellings, disk has preference over disc
        "disc":     ["NONE",       util.noop],
        "disk":     ["NONE",       util.noop],
        "mediadef": ["NONE",       util.noop],
        "nbbc":     [8.,           util.noop],
        "dbbcver":  ["ASTRO",      util.noop],
        "hor_az":   [[],           util.noop],
        "hor_el":   [[],           util.noop],
        "ax1lim":   [[-90., 450.], util.noop],
        "ax2lim":   [[2., 90.],    util.noop],
        "ax1rate":  [1000.,        util.noop],
        "ax2rate":  [1000.,        util.noop],
        "ax1acc":   [[1000., 0],   util.noop],
        "ax2acc":   [[1000., 0],   util.noop],
        "mount":    ["ALTAZ",      util.noop],
        "axistype": [0.,           util.noop],
        "axisoff":  [0.,           util.noop],
        "tsettle":  [0.,           util.noop],
        "minsetup": [0.,           util.noop],
        "tscal":    ["GAP",        util.upper],
        "maxsrchr": [1e6,          util.noop],
        "tlevset":  [0.,           util.noop],
        "endcat":   [0.,           util.noop],
    }
    input_iterator.set_defaults(station_record_defaults, station_state_defaults)
    
    location_state_defaults = {
        "version": ["Not known", util.noop],
    }
    
    location_record_defaults = {
        "dbname":   ["",  util.noop],
        "dbcode":   ["",  util.noop],
        "frame":    ["",  util.noop],
        "axistype": ["",  util.noop],
        "axisoff":  [0.,  util.noop],
        "x":        [0.,  util.noop],
        "y":        [0.,  util.noop],
        "z":        [0.,  util.noop],
        "dxdt":     [0.,  util.noop],
        "dydt":     [0.,  util.noop],
        "dzdt":     [0.,  util.noop],
        "epoch":    [0.,  util.noop],
        "begin":    [0.,  util.noop],
        "end":      [1e5, util.noop],
    }
    
    locations = collections.defaultdict(list)
    if not noloc:
        for loc in location_keyin_data:
            values, present = util.merge_record_with_defaults(
                loc, location_record_defaults, location_state_defaults)
            locations[values["dbname"]].append(values)
    
    # warn about location close together with time overlap
    # don't warn about known cases
    compare_locations = [loc for loc in sum(locations.values(), []) 
                         if loc["dbname"] not in ("MIAMI20", "SINTOTU") and 
                         not loc["dbname"].startswith("VLA")]
    for loc_a, loc_b in itertools.combinations(compare_locations, 2):
        separation = math.sqrt(
            sum([(loc_a[d] - loc_b[d]) ** 2 for d in "xyz"]))
        time_overlap = (min(loc_a["end"], loc_b["end"]) >
                        max(loc_a["begin"], loc_b["begin"]))
        if (separation < 10.) and time_overlap:
            s.wlog(0, "RDLOC:  Locations.dat stations {}   and {}   appear to "
                   "be the same (time and position).  Sep (m):{:10.2f}".format(
                       loc_a["dbname"], loc_b["dbname"], separation))
            
    # update the station values with station and location file catalog values
    station_catalog = StationCatalog()

    warnxyz = True
    locwarn = True
    # map from station catalog entry attributes to keyin keywords
    attribute_to_key = dict(zip(station_catalog.attributes, 
                                station_catalog.attributes))
    attribute_to_key.update({
        "long_bn": "long",
        "horaz": "hor_az",
        "horel": "hor_el",
        "xpos": "x",
        "ypos": "y",
        "zpos": "z",
        "dxpos": "dxdt",
        "dypos": "dydt",
        "dzpos": "dzdt",
        "mjdrate": "epoch",
        "axoff": "axisoff",
        "stndriv": "ndrives"})
    
    start = int(s.schsta.msta)
    index = start
    seen = set(entry.station for entry in station_catalog.entries[:index])
    for record in input_iterator:
        station_values, present = util.merge_record_with_defaults(
            record, station_record_defaults, station_state_defaults)
        if "endcat" in present:
            break

        if index >= len(station_catalog.entries):
            s.errlog(" STREAD: Too many stations in catalog, max {} "
                     " Last station: {}".format(
                         len(station_catalog.entries), 
                         station_catalog.entries[-1].station))

        if station_values["station"] in seen:
            s.wlog(0, "STREAD:  Ignoring extra station catalog entry for {}".\
                   format(station_values["station"]))
            continue
        else:
            seen.add(station_values["station"])
            

        # disc/disk confusion cases
        if "disk" not in present:
            station_values["disk"] = station_values["disc"]
        if station_values["mediadef"] == "DISC":
            station_values["mediadef"] = "DISK"

        # copy all values to the common block placeholder
        entry = station_catalog.entries[index]
        entry.set_keyin_values(station_values, attribute_to_key)

        entry.stcodeu = entry.stcode.upper()
        
        # fill in x, y, z from elev, lat, long (or reverse)
        gotxyz = ((entry.xpos, entry.ypos, entry.zpos) != (0., 0., 0.))
        gotllh = ((entry.elev, entry.lat, entry.long_bn) != (0., 0., 0.))
        if gotxyz and gotllh and warnxyz:
            s.putout("RDSTA: Both XYZ and Lat/Long coordinates given for some "
                     "stations.")
            s.putout("       There is nothing to insure they agree.")
            warnxyz = False
        
        conversion_mode = None # 0: llh -> xyz, 1: xyz -> llh
        if gotxyz and (not gotllh or (entry.elev > 1e6)):
            conversion_mode = 1
        elif gotllh and not gotxyz:
            conversion_mode = 0
        elif not gotllh and not gotxyz:
            if noloc or (len(locations) == 0):
                if locwarn:
                    if locfile.upper() == "NOLOC":
                        s.putout("RDSTA: **Cannot use locations catalog with "
                                 "in-line stations catalog.")
                    elif locfile.upper() != "NONE":
                        s.putout("RDSTA:  Could not open locations catalog: "
                                 "{}".format(locfile))
                    locwarn = False
                s.putout("RDSTA:   No coordinates for {}".format(
                    station_values["station"]))
            else:
                # get xyz from locations catalog
                station_locations = locations.get(station_values["dbname"])
                if station_locations is not None:
                    for location in station_locations:
                        if location["begin"] <= mjd1 < location["end"]:
                            for attribute in ["xpos", 
                                              "ypos",
                                              "zpos",
                                              "dxpos",
                                              "dypos",
                                              "dzpos",
                                              "mjdrate",
                                              "axoff"]:
                                setattr(entry, attribute, 
                                        location[attribute_to_key[attribute]])
                            conversion_mode = 1
                            break
                if conversion_mode is None:
                    s.putout("RDSTA: No coordinates in stations or locations "
                             "file for {} dbname: {}".format(
                                 station_values["station"], 
                                 station_values["dbname"]))
                    s.putout("RDSTA: Check date range in addition to names.")
                   
        if conversion_mode is not None:
            (entry.long_bn, entry.lat, entry.elev, 
             entry.xpos, entry.ypos, entry.zpos, ier) = \
                s.geoxyz(conversion_mode, 
                         entry.long_bn, entry.lat, entry.elev, 
                         entry.xpos, entry.ypos, entry.zpos)
            if ier != 0:
                s.putout("RDSTA: Problem with coordinate conversions for {}".\
                         format(station_values["station"]))
        
        if entry.ax1acc[1] == 0:
            entry.ax1acc[1] = entry.ax1acc[0]
        if entry.ax2acc[1] == 0:
            entry.ax2acc[1] = entry.ax2acc[0]

        entry.naxlim = min(len(entry.ax1lim), len(entry.ax2lim)) // 2

        if "zalim" not in present:
            if entry.mount == "ALTAZ":
                entry.zalim = 90. - entry.ax2lim[0]
            else:
                entry.zalim = 90.

        control = station_values["control"]
        entry.vlbadar = ((len(control) >= 5) and (control[4] == 'V'))
        
        dar = station_values["dar"]
        entry.useonsrc = ((dar.startswith("RDBE") and 
                           control.startswith("VLBA")) or 
                          ((dar == "WIDAR" ) and 
                           (control == "VEX")))

        def check(value, allowed, not_supported, station, element_type):
            if value not in allowed:
                s.errlog("STREAD: Invalid {} type {} for {}".format(
                    element_type, value, station))
            if value in not_supported:
                s.errlog("STREAD: {} type {} no longer supported by "
                         "SCHED.  Station: {}".format(
                             element_type[0].upper() + element_type[1:], 
                             value, station))

        check(entry.control, 
              ("VLA", "VLBA", "NRAO", "NRAOV", "SNAP", "VEX", "SN50", "VSOP", 
               "NONE"),
              ("SNAP", "SN50"),
              entry.station,
              "control")
        if entry.control.startswith("VLA"):
            s.errlog("STREAD: Control type VLA (old system card images) no "
                     "longer supported.  Use VEX.")

        check(entry.dar,
              ("VLBA", "RDBE", "RDBE2", "DBBC", "DBBC3", "VLBAG", 
               "MKIV", "MKIII", "S2", "K4", "K5", "VERA", "VSOP", 
               "VLBA4", "LBA", "R1002", "WIDAR", "CDAS", "eMERL", 
               "NONE"),
              ("MKIII", "S2"),
              entry.station,
              "DAR")

        check(entry.recorder,
              ("VLBA", "MKIV", "VLBA4", "MKIII", 
               "S2", "K4", "K5", "VERA", "VSOP", 
               "MARK5A", "MARK5B", "MARK5C", "NONE"),
              ("MKIII", "S2"),
              entry.station,
              "recorder")
            
        check(entry.disk,
              ("MARK5A", "MARK5B", "MARK5C", "LBADR", "NONE"),
              tuple(),
              entry.station,
              "DISK")

        check(entry.mediadef,
              ("TAPE", "DISK", "NONE"),
              tuple(),
              entry.station,
              "MEDIADEF")

        if (entry.xpos, entry.ypos, entry.zpos) == (0., 0., 0.):
            s.errlog("STREAD: Location required for {}: infile={}, locfile={}".\
                     format(entry.station, infile, locfile))

        if entry.tscal not in ("CONT", "GAP"):
            s.errlog("STREAD: Unknown TSCAL for {}:  {}".format(
                entry.station, entry.tscal))
        
        index += 1
              
    s.schsta.msta = index

    s.schcst.stver = util.resize_string(station_state_defaults["version"][0],
                                        s.schcst.stver.itemsize, "version")
    s.schcst.locaver = util.resize_string(location_state_defaults["version"][0],
                                          s.schcst.locaver.itemsize, "version")

    station_catalog.write(range(start, index))




from ..util import f2str
from ..catalog import StationCatalog, ScanCatalog, SetupCatalog
from . import defset

import schedlib as s

def defaults():
    if s.schcon.debug:
        s.wlog(0, "DEFAULTS starting.")

    # Protect against mixing RDBE and non-RDBE stations and frequency files.
    if (f2str(s.schsco.freqfile).find("RDBE") != -1) != \
       (f2str(s.schcst.stafile ).find("RDBE") != -1):
        s.wlog(1, " ")
        s.wlog(1, "DEFAULTS: **** Did you intend to mix RDBE and non-RDBE "
               "station and frequency files? ****")
        s.wlog(1, " ")

    # Discourage use of MARK5A on the VLBA except for pointing.
    station_catalog = StationCatalog()
    station_catalog.read()
    station_catalog.read_scheduled_attributes()
    if not s.schn1.notape and any(station.station.startswith("VLBA") and 
                                  (station.disk == "MARK5A") 
                                  for station in station_catalog.used()):
        s.wlog(1, "STREAD: ==== WARNING ==== Mark5A specified for a VLBA "
               " station.")
        s.wlog(1, "        Those recorders have been removed from most "
               "VLBA stations.")

    s.srfinish()
    s.pkfinish()
    s.vlascns()

    if not s.schcon.noset:
        defset()

        s.sdopincr()

    s.recctl()

    # Default the grab stuff.  Put after setups defaulted so we
    # have the bit rate.  The GRABGAP assumes that the required
    # bit rate is the total over all channels.
    scan_catalog = ScanCatalog()
    scan_catalog.read()
    setups = SetupCatalog().read()
    used_scans = scan_catalog.used()
    for scan_index, scan in enumerate(used_scans, scan_catalog.scan_offset):
        if (scan.grabto != "NONE") and (scan.datapath == "IN2DISK"):
            if scan.grabtime[0] < 0:
                scan.grabtime[0] = 30
            if scan.grabtime[1] < 0:
                scan.grabtime[1] = 10
            if scan.grabgap == 0:
                for station in station_catalog.used():
                    if station.stascn[scan_index]:
                        bps = setups[station.nsetup[scan_index] - 1].totbps
                        scan.grabgap = max(scan.grabgap,
                                           5 + scan.grabtime[0] * bps / 110.)
    scan_catalog.write(range(scan_catalog.scan_offset,
                             scan_catalog.scan_offset + len(used_scans)))
    



from . import scndup
from ..catalog import ScanCatalog, StationCatalog
from ..util import f2str

import schedlib as s

import numpy as np

import math
import itertools

max_input_scans = 1000
cell_time = np.empty(dtype=float, shape=(3, 3, StationCatalog.maxsta))
last_j_scan = 0
az_coff = 60
el_cell = [20, 45]
az_cell = [120, 240]
def optcells(last_scan_index, k_scan, scan_index):
    global cell_time, last_j_scan, weight_station

    if s.schcon.debug:
        s.wlog(1, "Starting OPTCELLS")

    done = False
    keep = True
    adjust = True
    
    stations = StationCatalog().used(use_direct_access=True)
    scans = ScanCatalog().direct_access_entries
    
    if k_scan == 1:
        if s.schn1.nscans > max_input_scans:
            s.errlog(" OPTCELLS:  Too many input scans for cell optimizer "
                     "- max: {}".format(max_input_scans))
        
        if s.schcon.opdur <= 0:
            s.errlog(" OPTCELLS: For OPTMODE=CELLS, OPDUR must be given.")

        adjust = False
        scans[scan_index - 1].startj = scans[0].startj
        approx_time = scans[scan_index - 1].startj
        last_j_scan = 0

        cell_time.fill(scans[0].startj - 0.5 / 24)
        baseline = np.zeros(dtype=float, shape=(len(stations), len(stations)))
        for (i, station1), (j, station2) in itertools.combinations(
                enumerate(stations), 2):
            baseline[i,j] = math.sqrt((station1.xpos - station2.xpos) ** 2 +
                                      (station1.ypos - station2.ypos) ** 2 +
                                      (station1.zpos - station2.zpos) ** 2)
        baseline += baseline.T
        mean_base = np.sum(baseline) / (len(stations) * (len(stations) - 1))
        weight_station = np.square(np.sum(baseline, axis=1) / 
                                   ((len(stations) - 1) * mean_base))
    else:
        approx_time = scans[scan_index - 2].stopj
    
    got_station = np.zeros(dtype=int, shape=(s.schn1.nscans,))
    ok_station = np.full(fill_value=False, dtype=bool, 
                         shape=(s.schn1.nscans, len(stations)))
    el_cell_index = np.empty(dtype=int, shape=(s.schn1.nscans, len(stations)))
    az_cell_index = np.empty(dtype=int, shape=(s.schn1.nscans, len(stations)))
    time_source = np.empty(dtype=float, shape=(s.schn1.nscans,))
    points = np.zeros(dtype=float, shape=(s.schn1.nscans,))
    for j_scan, scan in enumerate(scans[:s.schn1.nscans]):
        if j_scan + 1 != last_j_scan:
            for i, station in enumerate(stations):
                if station.stascn[j_scan]:
                    last_time, station.tonsrc[j_scan] = s.stageo(
                        j_scan + 1, i + 1, approx_time, last_scan_index[i],
                        "OPTCELLS")
                    if (f2str(station.up1[j_scan]) == "") and \
                       (f2str(station.up2[j_scan]) == "") and \
                       (station.el1[j_scan] > scan.opminel) and \
                       (station.el2[j_scan] > scan.opminel):
                        ok_station[j_scan, i] = True
                        got_station[j_scan] += 1

                        az_test = (station.az1[j_scan] + az_coff) % 360
                        
                        if station.el1[j_scan] < el_cell[0]:
                            el_cell_index[j_scan, i] = 1
                        elif station.el1[j_scan] < el_cell[1]:
                            el_cell_index[j_scan, i] = 2
                        else:
                            el_cell_index[j_scan, i] = 3

                        if el_cell_index[j_scan, i] == 3:
                            az_cell_index[j_scan, i] = 1
                        else:
                            if az_test < az_cell[0]:
                                az_cell_index[j_scan, i] = 1
                            elif az_test < az_cell[1]:
                                az_cell_index[j_scan, i] = 2
                            else:
                                az_cell_index[j_scan, i] = 3
            
            time_source[j_scan] = 1e10                
            all_0 = True
            for i, station in enumerate(stations):
                if ok_station[j_scan, i] and (last_scan_index[i] != 0):
                    all_0 = False
                    time_source[j_scan] = min(time_source[j_scan], 
                                              station.tonsrc[j_scan])
                    if s.schcon.opnosub and (scan_index > s.schn1.scan1):
                        time_source[j_scan] = max(time_source[j_scan], 
                                                  scans[scan_index - 2].stopj)
            if all_0:
                time_source[j_scan] = scans[0].startj

            if got_station[j_scan] >= scan.opmian:
                for i, station in enumerate(stations):
                    if ok_station[j_scan, i]:
                        eci = el_cell_index[j_scan, i]
                        aci = az_cell_index[j_scan, i]
                        
                        weight_multiplier = 1
                        if eci == 1:
                            for iaz in range(1, 4):
                                if iaz != aci:
                                    pt_add = 1440 * \
                                             (time_source[j_scan] - 
                                              cell_time[eci - 1, iaz - 1, i])
                                    weight_multiplier += pt_add ** 2 / \
                                        (s.schcon.optlowt ** 2 + pt_add ** 2)

                        lc_time = cell_time[eci - 1, aci - 1, i]
                        
                        pt_add = 1440 * (time_source[j_scan] - lc_time)
                        points[j_scan] += weight_station[i] * pt_add * \
                                          weight_multiplier
    
    max_points = np.max(points)

    time1 = 1e10
    for j_scan, scan in enumerate(scans[:s.schn1.nscans]):
        if got_station[j_scan] >= scan.opmian:
            time1 = min(time1, time_source[j_scan])

    for j_scan, scan in enumerate(scans[:s.schn1.nscans]):
        if got_station[j_scan] >= scan.opmian:
            sl_adj = 1440 * (time_source[j_scan] - time1)
            weight_multiplier = s.schcon.optslew ** 2 / (sl_adj ** 2 +
                                                         s.schcon.optslew ** 2)
            points *= weight_multiplier

    look_back = 6
    station_skip = np.zeros(dtype=int, shape=(len(stations),))
    if scan_index > s.schn1.scan1 + 2:
        i_skip1 = max(scan_index - look_back, s.schn1.scan1)
        for l, l_scan in enumerate(scans[i_skip1 - 1: scan_index - 1]):
            for i, station in enumerate(stations):
                if not (station.stascn[l] and (station.el1[l] > l_scan.opminel)):
                    station_skip[i] += 1

    for j, scan in enumerate(scans[:s.schn1.nscans]):
        max_skip = 0
        for i, station in enumerate(stations):
            if not (station.stascn[j_scan] and 
                    (station.el1[j_scan] > scan.opminel)):
                max_skip = max(max_skip, station_skip[i])

        if max_skip > look_back // 3:
            points[j_scan] *= 0.2

    max_scan = np.argmax(points)
    max_apt = points[max_scan]
    if max_apt > -1e10:
        j_scan = max_scan + 1
    else:
        s.errlog("OPTCELLS: Something weird went wrong with POINTS.")

    scndup(scan_index - 1, j_scan - 1, False, "OPTCELLS")
    scan = scans[scan_index - 1]
    scan.startj = time_source[j_scan - 1]
    scan.stopj = scan.startj + scan.dur

    last_j_scan = j_scan

    year, day, time_rad = s.timej(scan.startj)
    time_text = f2str(s.tformwrp(time_rad, "T", 0, 2, 2, "::@"))
    s.wlog(0, " Source, time, maxpts, maxapt: {:<12} {:>4d} {} "
           "{:>9.3f} {:>9.3f}".format(
               scan.scnsrc, day, time_text, max_points, max_apt))

    for i, station in enumerate(stations):
        if ok_station[j_scan - 1, i]:
            cell_time[el_cell_index[j_scan - 1, i] - 1, 
                      az_cell_index[j_scan - 1, i] - 1,
                      i] = scan.startj
            station.stascn[scan_index - 1] = True
        else:
            station.stascn[scan_index - 1] = False

    return adjust, keep, done



from .parameter import max_seg, secpday
from . import makeseg, geochk
from ..util import f2str_array, f2str

import schedlib as s

import numpy as np

dummy_index = -12345

def geomake(last_scan_index, j_scan, scan_index, n_seg, geo_stascn, geo_startj,
            stations, scans):
    global dummy_index

    sigma = np.empty(shape=(len(stations) * 2,), dtype=float)
    best_sigma = np.empty(shape=(len(stations),), dtype=float)
    if s.schcon.debug or (s.schsou.geoprt >= 2):
        s.wlog(1, "GEOMAKE starting.")

    n_print_stations = min(20, len(stations))

    seg_sources = np.zeros(shape=(max_seg,), dtype=int)
    
    kept = False

    for _ in range(100):
        dummy_index, _ = s.ran5wrap(dummy_index)
    
    start_time = 0
    for station_index, station in enumerate(stations):
        if station.stascn[j_scan - 1]:
            last_station_scan_index = last_scan_index[station_index]
            if last_station_scan_index != 0:
                start_time = max(start_time, 
                                 scans[last_station_scan_index - 1].stopj)
            else:
                start_time = max(start_time, s.schn1.tfirst)

    scan = scans[scan_index - 1]
    end_time = start_time + scan.geolen
    if s.schsou.geoprt >= 2:
        s.wlog(0, "GEOMAKE START TIMES {} {} {}".format(
            start_time, scan.startj, end_time))

    scan = scans[j_scan - 1]
    if scan.opmian < 2:
        s.wlog(1, "GEOMAKE:  OPMINANT of {} too small for automatic insertion "
               "of geodetic sections.".format(scan.opmian))
        scan.opmian = len(stations) // 2
        s.wlog(1, "          Resetting to OPMINANT = {}".format(scan.opmian))
    
    ok_geo, use_geo, seg_elevation = geochk(
        j_scan, scan_index, start_time, end_time)
    
    best_quality = 9999
    for trial in range(1, s.schsou.geotries + 1):
        if s.schsou.geoprt >= 1:
            s.wlog(0, "GEOMAKE starting to construct trial segment {}.  "
                   "First scan: {}".format(trial, scan_index))

        last_scan, n_scan_seg, test_seg_sources, dummy_index, selection_type, \
            worst_station = makeseg(
                j_scan, scan_index, last_scan_index, ok_geo, use_geo, 
                seg_elevation, start_time, end_time, dummy_index, sigma,
                stations, scans)
        min_scans = min(np.count_nonzero(
            station.stascn[scan_index - 1: last_scan])
                        for station in stations)

        retain = (min_scans >= 3)

        if retain:
            if s.schsou.geoprt >= 2:
                n_prt = min(20, n_scan_seg)
                if n_scan_seg > n_prt:
                    print(" Showing first {} scans of {}".format(
                        n_prt, n_scan_seg))
                s.wlog(0, "GEOMAKE ---------------------------------- ")
                s.wlog(0, "GEOMAKE    FINISHED ONE SEQUENCE - Trial: {}  "
                       "Geosrcs: {}".format(trial, " ".join(
                           "{:4d}".format(s) for s in test_seg_sources[:n_prt]
                       )))

            print_debug = (s.schsou.geoprt >= 2)
            test_quality, sigma = s.geoqual(scan_index, scan_index, last_scan, 
                                            j_scan, 14.48, print_debug, sigma)

            if s.schsou.geoprt >= 0:
                s.wlog(0, " ")
                s.wlog(0, "GEOMAKE: Finished making trial segment. "
                       "{} Number of scans: {}  Quality: {:.2f}  "
                       "Previous best: {:.2f}".format(
                           trial, n_scan_seg, test_quality, best_quality))

                if n_scan_seg > 50:
                    s.wlog(1, "  Only first 50 scans printed")
                    n_prt = 50
                else:
                    n_prt = n_scan_seg
                
                s.wlog(0, "     Sources:   {}".format(" ".join(
                    "{:3d}".format(s) for s in test_seg_sources[:n_prt])))
                s.wlog(0, "     Priority:  {}".format(" ".join(
                    "{:3d}".format(use_geo[s - 1]) 
                    for s in test_seg_sources[:n_prt])))
                s.wlog(0, "     Sigmas by station: {}".format(" ".join(
                    "{:.1f}".format(s) for s in sigma[:n_print_stations])))

            if test_quality < best_quality:
                if s.schsou.geoprt == 0:
                    s.wlog(0, " ")
                
                s.wlog(0, "New best geodetic segment - Trial: {} "
                       "Number of scans & fewest scans/sta: {} {}  "
                       "Quality: {:.3f}".format(
                           trial, n_scan_seg, min_scans, test_quality))

                best_quality = test_quality
                seg_sources[:n_scan_seg] = test_seg_sources[:n_scan_seg]
                geo_startj[:n_scan_seg] = \
                    [scan.startj for scan in scans[scan_index - 1: 
                                                   scan_index + n_scan_seg - 1]]
                best_sigma[:] = sigma[:best_sigma.shape[0]]
                for station_index, station in enumerate(stations):
                    geo_stascn[:n_scan_seg, station_index] = station.stascn[
                        scan_index - 1: scan_index + n_scan_seg - 1]

                n_seg = n_scan_seg
                kept = True
                
        else:
            if s.schsou.geoprt >= 0:
                s.wlog(0, "GEOMAKE: Segment not kept.  "
                       "Not enough scans at some stations.")


        if ((s.schsou.geoprt >= 0) and kept) or (s.schsou.geoprt >= 1):
            if retain:
                s.wlog(0, "   Sigmas by station:                {}".format(
                    " ".join("{:.2f}".format(s) for s in sigma[:n_print_stations]
                         )))
            else:
                s.wlog(0, "  Rejected without testing.  "
                       "Some station(s) with too few scans.")

            s.wlog(0, " Num  Gap(s) Sel Source      Worst {}".format(
                "".join("{:^7}".format(station.stcode) 
                        for station in stations[:n_print_stations])))
            for i in range(n_scan_seg):
                j = scan_index + i - 1
                if (i == 0) and (scan_index == s.schn1.scan1):
                    s_gap = 0
                else:
                    s_gap = (scans[j].startj - scans[j - 1].stopj) * secpday

                seg_source_index = test_seg_sources[i]
                source = f2str(s.schcsc.geosrc[seg_source_index - 1])
                elevation_text = " ".join(
                    " {:>4.0f} ".format(station.el1[j]) if station.stascn[j] 
                    else "({:>4.0f})".format(station.el1[j]) 
                    for station in stations[:n_print_stations])
                s.wlog(0, "{:4d} {:>7.0f} {:<3} {:<12}  {} {}".format(
                    test_seg_sources[i], s_gap, selection_type[i], source, 
                    worst_station[i], elevation_text))
                
        kept = False
        
        if trial % 10 == 0:
            s.wlog(1, "Finished trial geodetic segment number {}".format(trial))

    # end of trial loop
    
    if s.schsou.geoprt >= 0:
        n_prt = min(30, n_scan_seg)
        s.wlog(0, "   ")
        if n_scan_seg > n_prt:
            s.wlog(0, " Showing first {} scans of {}".format(n_prt, n_scan_seg))
        else:
            s.wlog(0, " ")
        
        s.wlog(0, "GEOMAKE FINISHED: {} Selected geodetic sources: {}".format(
            n_seg, " ".join("{:4d}".format(s) for s in seg_sources[:n_prt])))

    if s.schsou.geoprt < 0:
        s.wlog(0, "                                     " + 
               " ".join("{:^8}".format(station.stcode) 
                        for station in stations[:n_print_stations]))

    sigmas_text = " ".join("{:.2f}".format(s) 
                          for s in best_sigma[:n_print_stations])
    s.wlog(1, " Quality: {:.2f}  Sigmas by station: {}".format(
        best_quality, sigmas_text))

    if s.schsou.geoprt >= 0:
        stations_text = "".join("{:^7}".format(station.stcode) 
                                for station in stations[:n_print_stations])
        s.wlog(0, " Num   Gap(s)  Source    " + stations_text)

    return seg_sources, n_seg, geo_stascn, geo_startj



from . import ifdbbc, ifdbbc3
from ..catalog import SetupCatalog

import schedlib as s

import numpy as np

def chkdbbc(ks, setup_entry, station_entry):
    if s.schcon.debug:
        s.wlog(0, "CHKDBBC: Starting")
    
    errs = False
    
    if station_entry.dar == "DBBC3":
        ifnam = "ABCDEFGH"
        ifbbc, mif = ifdbbc3()
    else:
        ifnam = "ABCD"
        e_firmware = SetupCatalog.is_dbbc_e_firmware(setup_entry)
        ifbbc, mif = ifdbbc(station_entry.dbbcver, e_firmware)

    def check_ifchan():
        somebad = False
        for ifchan in setup_entry.ifchan:
            if (ifchan[0] not in ifnam) or \
               ((len(ifchan) > 1) and (ifchan[1] not in "1234 ")):
                if_descriptors = [c + "[1-4]" for c in ifnam]
                ifnam_text = ", ".join(if_descriptors[:-1]) + \
                             " or " + if_descriptors[-1]
                s.wlog(1, "CHKDBBC: IFCHAN '{}' not {}".format(
                    ifchan, ifnam_text))
                somebad = True
        return somebad

    def check_channels():
        somebad = False
        for ich, channel in enumerate(setup_entry.channel):
            iif = ifnam.find(channel.ifchan[0])
            bbc = channel.bbc
            if (iif == -1) or (not ifbbc[bbc-1, iif]):
                s.wlog(1, "CHKDBBC: Illegal IF input {} for DBBC, channel "
                       "{} BBC {}".format(channel.ifchan, ich+1, bbc))
                s.wlog(1, "         Allowed IF index and first character "
                       "for this BBC are: {}".format(
                           "".join("({}, {})".format(i+1, ifnam[i])
                                   for i in np.where(ifbbc[bbc-1] == 1)[0])
                       ))
                somebad = True
        if somebad:
            s.wlog(1, "        Be careful of special wiring restrictions "
                   "for these DARs.")
        return somebad

    if s.schn1.vlbitp and (setup_entry.format != "NONE") and \
       (len(np.unique(setup_entry.ifchan)) > mif):
        s.wlog(1, "CHKDBBC: More than {} IF's requested for DBBC station {}.".\
               format(mif, station_entry.station))
        errs = True

    if station_entry.dar == "DBBC3":
        errs = check_ifchan() or errs
        errs = check_channels() or errs

    elif setup_entry.dbe == "DBBC_PFB":
        if not setup_entry.modetest:
            s.wlog(1, "CHKDBBC: You have specified DBE=DBBC_PFB The PFB "
                   "mode is not yet supported in the DBBC. This schedule "
                   "will probably fail. Specify MODETEST if you are "
                   "testing.")
            errs = True

        if setup_entry.nchan != 16:
            s.wlog(1, "CHKDBBC: For DBE=DBBC_PFB, NCHAN must be 16. Setup "
                   "specified: {}".format(setup_entry.nchan))
            errs = True

        if setup_entry.samprate != 64:
            s.wlog(1, "CHKDBBC: Invalid SAMPRATE specified: {} for "
                   "DBE=DBBC_PFB. Must be 64.0 Msamp/s.")
            errs = True

        bits = set(setup_entry.bits)
        if bits <= {1,2}:
            s.wlog(1, "CHKDBBC: BITS must be 1 or 2 for DBE=DBBC_PFB. "
                   "Value specified is: {}".format(
                       ", ".join(str(i) for i in sorted(bits))))
            errs = True

        if len(np.unique(setup_entry.sidebd)) != 1:
            s.wlog(1, "CHKDBBC: SIDEBAND must be the same for all channels "
                   "if DBE=DBBC_PFB.")
            errs = True

    elif setup_entry.dbe == "DBBC_DDC":
        allowed_sample_rates = SetupCatalog.dbbc_firmware_allowed_sample_rates(
            setup_entry)
        if setup_entry.samprate not in allowed_sample_rates:
            firmware_text = ", firmware {}".format(setup_entry.dbbcfw) \
                            if setup_entry.dbbcfw != "" else ""
            s.wlog(1, "CHKDBBC: Invalid SAMPRATE specified: {} for "
                   "DBE=DBBC_DDC{}. Must be {} to {} Msamp/s.".format(
                       setup_entry.samprate, firmware_text,
                       min(allowed_sample_rates), max(allowed_sample_rates)))
            errs = True

        bits = set(b for b in setup_entry.bits)
        if not (bits <= {1,2}):
            s.wlog(1, "CHKDBBC: BITS must be 1 or 2 for DBE=DBBC_DDC. "
                   "Value specified is: {}".format(
                       ", ".join(str(i) for i in sorted(bits))))
            errs = True

        errs = check_ifchan() or errs
        errs = check_channels() or errs

    if station_entry.dar != "DBBC3":
        errs = s.chkdbfq(ks, setup_entry.bbfilt, setup_entry.bbsyn, errs, True)

    return errs
                



from . import scndup
from ..catalog import ScanCatalog, StationCatalog, SourceCatalog
from ..util import f2str

import schedlib as s

import numpy as np

skip = np.empty(dtype=int, shape=(SourceCatalog.maxsource))
missed = 0
def opthiel(last_scan_index, k_scan, scan_index):
    global skip, missed

    done = False
    keep = True
    adjust = True
    
    stations = StationCatalog().used(use_direct_access=True)
    scans = ScanCatalog().direct_access_entries
    
    if k_scan > s.schn1.nscans:
        done = True
        return k_scan, adjust, keep, done

    scan = scans[k_scan - 1]
    scan.higroup = max(scan.higroup, 1)
    if k_scan + scan.higroup - 1 > s.schn1.nscans:
        s.errlog("OPTHIEL: HIGROUP in last scan too large.")

    if scan.higroup == 1:
        scndup(scan_index - 1, k_scan - 1, True, "OPTHIEL")
        return k_scan, adjust, keep, done

    approx_time = s.schn1.tfirst
    for i, station in enumerate(stations):
        use_station = np.any(station.stascn[k_scan - 1: 
                                            k_scan - 1 + scan.higroup])
        if use_station and (last_scan_index[i] != 0):
            approx_time = max(approx_time, scans[last_scan_index[i] - 1].stopj +
                              scans[scan_index - 1].gap)

    group_max_low_el = -100
    for j in range(scan.higroup):
        j_k_scan = k_scan - 1 + j
        j_scan_index = scan_index - 1 + j
        scndup(j_scan_index, j_k_scan, True, "OPTHIEL")
        scan_low_el = 100
        for i, station in enumerate(stations):
            if station.stascn[j_scan_index]:
                s.stageo(j_scan_index + 1, i + 1, approx_time, 
                         last_scan_index[i], "OPTHIEL")
                average_el = (station.el1[j_scan_index] + 
                              station.el2[j_scan_index]) / 2
                scan_low_el = min(scan_low_el, average_el)
                
        if scan_low_el > group_max_low_el:
            group_max_low_el = scan_low_el
            max_scan = j_scan_index

    scndup(scan_index - 1, max_scan, True, "OPTHIEL")

    k_scan += scan.higroup - 1
    
    return k_scan, adjust, keep, done



from ..catalog import StationCatalog, ScanCatalog

import schedlib as s

import copy

station_catalog = StationCatalog()
scan_catalog = ScanCatalog()
copyall_attributes = {"startj", "stopj"}
def scndup(to, from_, copyall, caller, use_direct_access=True):
    if s.schcon.debug:
        s.wlog(0, "SCNDUP: Duplicating scan {} to scan {}.  Called by: {} "
               "copyall = {}".format(from_+1, to+1, caller, copyall))
        s.wlog(0, "SCNDUP: ANNOT: {}".format(f2str(s.schc2a.annot[from_])))
    if to >= ScanCatalog.maxscan:
        s.errlog("SCHDUP: Output scan number {} too big for arrays of "
                 "dimension: {}".format(to+1, ScanCatalog.maxscan))

    # copy the scan
    if use_direct_access:
        for block, items in scan_catalog.block_items.items():
            for attr in items:
                if copyall or (attr not in copyall_attributes):
                    array = getattr(block, attr)
                    array[..., to] = array[..., from_]
        entries = scan_catalog.entries
        for attr in scan_catalog.extended_attributes:
            setattr(entries[to], attr, 
                    copy.deepcopy(getattr(entries[from_], attr)))
    else:
        entries = scan_catalog.entries
        if not copyall:
            original = {k: getattr(entries[to], k) for k in copyall_attributes}
        # copy entry dict so the all references to entry are updated
        entries[to].__dict__ = copy.deepcopy(entries[from_].__dict__)
        if not copyall:
            for k, v in original.items():
                setattr(entries[to], k, v)
    
    s.schn2a.nsetup[to, :] = s.schn2a.nsetup[from_, :]
    s.schn2a.fseti[to, :] = s.schn2a.fseti[from_, :]
    s.schn2a.stascn[to, :] = s.schn2a.stascn[from_, :]
    s.schn2b.dopincr[to, :] = s.schn2b.dopincr[from_, :]
    if not copyall:
        # reset some entries
        if use_direct_access:
            s.schn2a.duronly[to] = 1
            s.schc2a.annot[to] = "".ljust(s.schc2a.annot.itemsize)
        else:
            entries[to].duronly = 1
            entries[to].annot = ""
    else:
        # copy station entries
        s.schn5.tpstart[to, :] = s.schn5.tpstart[from_, :]
        s.schn5.tcorr[to, :] = s.schn5.tcorr[from_, :]
        s.schn5.gbytes[to, :] = s.schn5.gbytes[from_, :]
        s.schn6.lst1[to, :] = s.schn6.lst1[from_, :]
        s.schn6.lst2[to, :] = s.schn6.lst2[from_, :]
        s.schn6.tonsrc[to, :] = s.schn6.tonsrc[from_, :]
        s.schn6.tslew[to, :] = s.schn6.tslew[from_, :]
        s.schn6.el1[to, :] = s.schn6.el1[from_, :]
        s.schn6.az1[to, :] = s.schn6.az1[from_, :]
        s.schn6.ha1[to, :] = s.schn6.ha1[from_, :]
        s.schn6.pa1[to, :] = s.schn6.pa1[from_, :]
        s.schc6.up1[to, :] = s.schc6.up1[from_, :]
        s.schn6.el2[to, :] = s.schn6.el2[from_, :]
        s.schn6.az2[to, :] = s.schn6.az2[from_, :]
        s.schn6.ha2[to, :] = s.schn6.ha2[from_, :]
        s.schn6.pa2[to, :] = s.schn6.pa2[from_, :]
        s.schc6.up2[to, :] = s.schc6.up2[from_, :]

    if not use_direct_access:
        # update the scan related attributes of stations in the catalog
        station_catalog.read_scheduled_attributes_for_scan(to)



from ..util import get_catalog_dir

import schedlib as s

import os.path

file_type_defaults = {
    "location": "locations.dat",
    "stations": "stations_RDBE.dat",
    "sources":  "sources.gsfc",
    "frequency": "freq_RDBE.dat",
    "peakcommand": "peak.cmd",
    "refpointing": "sources.peak",
    "messages": "messages.txt"}

def schdefs(file_type):
    if file_type not in file_type_defaults:
        s.putout("SCHFILES: Unrecognized file type: {}".format(file_type))
        s.error("Programming problem")
        return ""
    default = file_type_defaults[file_type]
    return os.path.join(get_catalog_dir(), default)



from .catalog import Catalog, map_attr_array
from ..util import f2str

import schedlib as s

class ScanCatalog(Catalog):
    """
    CHARACTER        CENTERS(MAXSCN)*12
    INTEGER          ICENT(MAXSCN)
    INTEGER          GEOISCN(MAXSCN)
    DOUBLE PRECISION GEOLEN(MAXSCN)
    INTEGER          ORIGEN(MAXSCN)
    LOGICAL          DODOWN(MAXSCN)
    REAL             OPMINEL(MAXSCN)
    REAL             OPSLEWWT(MAXSCN), OPSLEWTI(MAXSCN)
    REAL             OPHLIMWT(MAXSCN), OPHLIMTI(MAXSCN)
    REAL             OPHMAXDT(MAXSCN)
    REAL             OPMINSEP(MAXSCN), OPHA(MAXSCN)
    REAL             OPHAWID(MAXSCN), OPHAWT(MAXSCN)
    INTEGER          OPMIAN(MAXSCN)
    INTEGER          OPMISS(MAXSCN)
    DOUBLE PRECISION PRESCAN(MAXSCN), DUR(MAXSCN), GAP(MAXSCN)
    DOUBLE PRECISION MINDW(MAXSCN), STARTJ(MAXSCN), STOPJ(MAXSCN)
    DOUBLE PRECISION FREQ(MAXCHN,MAXSCN)
    DOUBLE PRECISION BW(MAXCHN,MAXSCN)
    DOUBLE PRECISION CRDFREQ(MAXCHN,MAXSCN), CRDBW(MAXCHN,MAXSCN)
    INTEGER          SRCNUM(MAXSCN), IDOPSRC(MAXSCN), QUAL(MAXSCN)
    INTEGER          SETNUM(MAXSCN)
    INTEGER          DOPEAK(MAXSCN)
    INTEGER          HIGROUP(MAXSCN)
    INTEGER          POINT(MAXSCN), DURONLY(MAXSCN), NOWAIT(MAXSCN)
    INTEGER          ISCINT(MSCINT,MAXSCN), NSCINT(MAXSCN)
    INTEGER          CRDNCH(MAXSCN), CRDCH1(MAXSCN)
    INTEGER          CRDSETCH(MAXCRD,MAXSCN)
    LOGICAL          GOTCRD(MAXSCN)
    LOGICAL          NOTSYS(MAXSCN), DOPCAL(MAXSCN), CRDDOP(MAXSCN)
    LOGICAL          DWELL(MAXSCN), TANT1(MAXSCN), TANT2(MAXSCN)
    CHARACTER        ANNOT(MAXSCN)*128, SCNSRC(MAXSCN)*12
    CHARACTER        DOPSRC(MAXSCN)*12, LINES(MAXSCN)*8
    CHARACTER        PCAL(MAXSCN)*4, SCANTAG(MAXSCN)*4
    CHARACTER        PREEMPT(MAXSCN)*5
    INTEGER          IVLAPHS(MAXSCN), VLAINTEG(MAXSCN)
    INTEGER          VLAPTIME(MAXSCN)
    LOGICAL          PHASING(MAXSCN)
    CHARACTER        VLATSYS(MAXSCN)*1, VLAPEAK(MAXSCN)*9
    CHARACTER        VLAMODE(MAXSCN)*2
    CHARACTER        VLAPHS(MAXSCN)*12
    INTEGER          PTSLEW(MAXSCN), CALTIME(MAXSCN)
    LOGICAL          PNTVLBA(MAXSCN), TANVLBA(MAXSCN), DOPN3DB(MAXSCN)
    REAL             FOCUS(MAXSCN), ROTATION(MAXSCN)
    REAL             SAZCOL(MAXSCN), SELCOL(MAXSCN)
    CHARACTER        CRDLINE(MAXSCN)*80
    DOUBLE PRECISION MINPAUSE(MAXSCN), PRESTART(MAXSCN)
    LOGICAL          NOREC(MAXSCN)
    REAL             GRABTIME(2,MAXSCN), GRABGAP(MAXSCN)
    CHARACTER        DATAPATH(MAXSCN)*8, GRABTO(MAXSCN)*4
    """
    maxscan = s.schn2b.prescan.shape[0]
    block_items = {
        s.schcsc: [
            "centers"
        ],
        s.schsou: [
            "icent",
            "geoiscn",
            "geolen"
        ],
        s.schcon: [
            "origen",
            "dodown",
            "opminel",
            "opslewwt",
            "opslewti",
            "ophlimwt",
            "ophlimti",
            "ophmaxdt",
            "opminsep",
            "opha",
            "ophawid",
            "ophawt",
            "opmian",
            "opmiss"
        ],
        s.schn2a: [
            "srcnum",
            "idopsrc",
            "qual",
            "setnum",
            "dopeak",
            "higroup",
            "point",
            "duronly",
            "nowait",
            "iscint",
            "nscint",
            "crdnch",
            "crdch1",
            "crdsetch",
            "gotcrd",
            "notsys",
            "dopcal",
            "crddop",
            "dwell",
            "tant1",
            "tant2"
        ],
        s.schn2b: [
            "prescan",
            "dur",
            "gap",
            "mindw",
            "freq"
        ],
        s.schn2c: [
            "startj",
            "stopj",
            "bw"
        ],
        s.schn2d: [
            "crdfreq",
            "crdbw"
        ],
        s.schc2a: [
            "annot"
        ],
        s.schc2b: [
            "scnsrc",
            "dopsrc",
            "lines",
            "pcal",
            "scantag",
            "preempt"
        ],
        s.schn3: [
            "ivlaphs",
            "vlainteg",
            "vlaptime",
            "phasing"
        ],
        s.schc3: [
            "vlatsys",
            "vlapeak",
            "vlamode",
            "vlaphs"
        ],
        s.schn4: [
            "ptslew",
            "caltime",
            "pntvlba",
            "tanvlba",
            "dopn3db",
            "focus",
            "rotation",
            "sazcol",
            "selcol",
        ],
        s.schc4: [
            "crdline"
        ],
        s.schn5: [
            "minpause",
            "prestart",
            "norec",
            "grabtime",
            "grabgap"
        ],
        s.schc5: [
            "datapath",
            "grabto"
        ]
    }

    class DirectAccessCatalogEntry(object):
        # a catalog entry which maps setattr and getattr directly to
        # the COMMON blocks in schedlib

        def __init__(self, index, attr_array):
            self.__dict__["_index"] = index
            self.__dict__["_attr_array"] = attr_array

        def __getattr__(self, attr):
            try:
                array = self._attr_array[attr]
            except KeyError:
                raise AttributeError("'{}' object has no attribute '{}'".\
                                     format(type(self).__name__, attr))

            if len(array.shape) > 1:
                return array[..., self._index]
            else:
                ret = array[self._index]
                if array.dtype.kind == "S":
                    return f2str(ret)
                return ret

        def __setattr__(self, attr, value):
            try:
                array = self._attr_array[attr]
            except KeyError:
                return super().__setattr__(attr, value)
            
            array[..., self._index] = value


    def __init__(self):
        super().__init__(self.maxscan, self.block_items, {"scanexps"})
        self.attr_array = map_attr_array(self.block_items)
        self.direct_access_entries = [
            self.DirectAccessCatalogEntry(i, self.attr_array) 
            for i in range(self.maxscan)]

    def read(self, selection_slice=None):
        self.scan_offset = s.schn1.scan1 - 1
        return super().read(selection_slice)

    def scheduled_slice(self):
        return slice(s.schn1.scanl)

    def used(self):
        return self.entries[self.scan_offset:s.schn1.scanl]



from .parameter import secpday, onesec
from . import scndup
from ..catalog import StationCatalog
from ..util import f2str, bool2str

import schedlib as s

import numpy as np

def addpeak(last_scan_index, scans, stations, setups, peak_groups, 
            peak_opt, insert_adjust, scan_index):
    if s.schcon.debug:
        s.wlog(0, "ADDPEAK starting")

    scan = scans[scan_index - 1]

    if (not s.schcon.autopeak) or ((scan.point >= -1) and (peak_opt == 0)):
        return peak_opt, insert_adjust

    assert(s.schpeakn.npkgrp > 0), "ADDPEAK: Want automatic pointing, but "\
        "there are no groups"

    t_gap = np.empty(dtype=float, shape=(len(stations),))
    try_group = np.full(fill_value=False, shape=(len(peak_groups),))
    try_group2 = np.full(fill_value=False, shape=(len(peak_groups),))
    # last_*_scan objects are passed to FORTRAN, so require a fixed size,
    # only len(stations) is actually used
    last_target_scan = np.empty(dtype=int, shape=(StationCatalog.maxsta,))
    last_pointing_scan = np.empty(dtype=int, shape=(StationCatalog.maxsta,))
    sr_ok = np.empty(dtype=bool, shape=(s.schpeakn.pksrnum.shape[0],))
    sr_ok2 = np.empty(dtype=bool, shape=(s.schpeakn.pksrnum.shape[0],))
    slew_max = np.empty(dtype=float, shape=(s.schpeakn.pksrnum.shape[0],))
    slew_min = np.empty(dtype=float, shape=(s.schpeakn.pksrnum.shape[0],))
    if peak_opt == 0:
        if s.schcon.pkwatch:
            s.wlog(0, " ")
            year, day1, start = s.timej(scan.startj)
            stime = f2str(s.tformwrp(start, "T", 0, 2, 2, "::@"))
            s.wlog(0, "ADDPEAK: Trying to add pointing scan for target scan "
                   "{} {} {}".format(scan_index, scan.scnsrc, stime))

        try_group = np.full(shape=(s.schpeakn.npkgrp,), fill_value=False)
        try_group2 = np.full(shape=(s.schpeakn.npkgrp,), fill_value=False)
        n_up = np.full(shape=(s.schpeakn.npkgrp,), fill_value=0)
        m_settle = np.full(shape=(s.schpeakn.npkgrp,), fill_value=0.0)
        
        n_good = 0
        for station in stations:
            group_index = station.pkgroup
            if station.stascn[scan_index - 1] and (group_index != 0) and \
               ((f2str(station.up1[scan_index - 1]) == "") or 
                (f2str(station.up2[scan_index - 1]) == "")):
                n_good += 1
                n_up[group_index - 1] += 1

        if n_good == 0:
            if s.schcon.pkwatch:
                s.wlog(0, "ADDPEAK:  Source down or below OPMINEL at all "
                       "stations.")
            return peak_opt, insert_adjust

        for station_index, station in enumerate(stations):
            if station.stascn[scan_index - 1] and (station.pkgroup > 0):
                setup = setups[station.nsetup[scan_index - 1] - 1]
                group_index = station.pkgroup
                freq_test = setup.freqref[0]
                peak_group = peak_groups[group_index - 1]
                if (peak_group.pkminfq < freq_test):
                    if (last_scan_index[station_index] != 0):
                        last_scan = scans[last_scan_index[station_index] - 1]
                        t_gap[station_index] = scan.startj - last_scan.stopj
                        lastptg = (last_scan.point in {0, group_index}) and \
                                  (t_gap[station_index] < 600. / secpday)
                        if (not lastptg) and (n_up[group_index - 1] > 0):
                            if t_gap[station_index] > peak_group.pkdwell:
                                try_group[group_index - 1] = True
                            if t_gap[station_index] > 2 * peak_group.pkdwell:
                                try_group2[group_index - 1] = True
                    else:
                        t_gap[station_index] = 10000
                        try_group[group_index - 1] = n_up[group_index - 1] > 0
                        try_group2[group_index - 1] = n_up[group_index - 1] > 0

            last_target_scan[station_index] = last_scan_index[station_index]
            last_pointing_scan[station_index] = last_scan_index[station_index]

        n_try = np.count_nonzero(try_group) + np.count_nonzero(try_group2)
        
        if s.schcon.pkwatch:
            if n_try >= 1:
                for group_index, peak_group in enumerate(peak_groups):
                    if try_group2[group_index]:
                        s.wlog(0, "ADDPEAK: Pointing group {}:  Will try to "
                               "add double pointing scans.".format(
                                   group_index + 1))
                    elif try_group[group_index]:
                        s.wlog(0, "ADDPEAK: Pointing group {}:  Will try to "
                               "add one pointing scan.".format(
                                   group_index + 1))
                    else:
                        if n_up[group_index - 1] > 0:
                            s.wlog(0, "ADDPEAK: Pointing group {}:  No "
                                   "stations up in scan.".format(
                                       group_index + 1))
                        else:
                            s.wlog(0, "ADDPEAK: Pointing group {}:  Gap too "
                                   "short to add pointing scans.".format(
                                       group_index + 1))
            else:
                s.wlog(0, "ADDPEAK: Insufficient time to add pointing or "
                       "last scan was pointing.")

        if n_try == 0:
            return peak_opt, insert_adjust

        keep_origen = scan.origen
        scan.origen = 4
        for k_scan in range(scan_index, scan_index + n_try):
            scndup(k_scan, scan_index - 1, True, "ADDPEAK")

        n_added = 0
        m_scan = scan_index + n_try
        for group_index, peak_group in enumerate(peak_groups, 1):
            if try_group[group_index - 1]:
                if s.schcon.pkwatch:
                    s.wlog(0, "ADDPEAK:  Working on pointing group {}".format(
                        group_index))
                    tab_head = True

                k_scan = scan_index + n_added
                got_station = False
                for station in stations:
                    station.stascn[k_scan - 1] = \
                        station.stascn[m_scan - 1] and \
                        (station.pkgroup == group_index) and \
                        ((f2str(station.up1[m_scan - 1]) == "") or 
                         (f2str(station.up2[m_scan - 1]) == ""))
                    if station.stascn[k_scan - 1]:
                        got_station = True

                if got_station:
                    for p_src in range(1, peak_group.npksrc + 1):
                        stop_k = scans[m_scan - 1].startj - 20.0 * onesec
                        start_k = stop_k - peak_group.pkdwell
                        k_src = peak_group.pksrnum[p_src - 1]
                        s.srinsert(k_scan, k_src, peak_group.pksrc[p_src - 1],
                                   start_k, stop_k, last_scan_index, m_scan)

                        sr_ok[p_src - 1] = True
                        sr_ok2[p_src - 1] = True
                        watch_it = False
                        min_el = 100
                        max_el = 0
                        min_slew = 1000
                        max_slew = 0
                        min_tot = 10000
                        for station in stations:
                            if station.stascn[k_scan - 1] and \
                               (station.el1[k_scan - 1] < peak_group.pkminel):
                                sr_ok[p_src - 1] = False
                                sr_ok2[p_src - 1] = False

                            if s.schcon.pkwatch:
                                if station.stascn[k_scan - 1] and \
                                   (f2str(station.up1[k_scan - 1]) == "") and \
                                   (f2str(station.up2[k_scan - 1]) == ""):
                                    watch_it = True

                                if station.stascn[k_scan -1]:
                                    min_el = min(min_el, 
                                                 station.el1[k_scan - 1])
                                    max_el = max(max_el,
                                                 station.el1[k_scan - 1])
                                    min_slew = min(
                                        min_slew, 
                                        station.tslew[m_scan - 1] * secpday)
                                    max_slew = max(
                                        max_slew,
                                        station.tslew[m_scan - 1] * secpday)
                    
                        if sr_ok[p_src - 1]:
                            for station_index, station in enumerate(stations):
                                if station.stascn[k_scan - 1]:
                                    tot_time = peak_group.pkdwell + \
                                        station.tslew[k_scan - 1] + \
                                        station.tslew[m_scan - 1]
                                    if t_gap[station_index] < tot_time:
                                        sr_ok[p_src - 1] = False
                                    min_tot = min(min_tot, tot_time * secpday)

                                    tot_time2 = 2 * peak_group.pkdwell + \
                                                station.tslew[k_scan - 1] + \
                                                station.tslew[m_scan - 1] + \
                                                station.tsettle / secpday
                                    if t_gap[station_index] < tot_time2:
                                        sr_ok2[p_src - 1] = False

                                    m_settle[group_index - 1] = \
                                        max(m_settle[group_index - 1],
                                            station.tsettle)
                        
                        slew_max[p_src - 1] = 0
                        if sr_ok[p_src - 1]:
                            for station in stations:
                                if station.stascn[k_scan - 1]:
                                    slew_max[p_src - 1] = max(
                                        slew_max[p_src - 1], 
                                        station.tslew[m_scan - 1])

                        if s.schcon.pkwatch and watch_it:
                            if tab_head:
                                s.wlog(0, " Num   Source      Slew times (sec) "
                                       "Elevations (deg)  Scans OK  Min T")
                                tab_head = False

                            s.wlog(0, "{:>6} {:<12}"
                                   "{:>6.0f} to {:<6.0f} "
                                   "{:>6.0f} to {:<6.0f}"
                                   "    {} {}   {:>7.0f}".format(
                                       p_src, peak_group.pksrc[p_src - 1], 
                                       min_slew, max_slew, min_el, max_el, 
                                       bool2str(sr_ok[p_src - 1]), 
                                       bool2str(sr_ok2[p_src - 1]), 
                                       float(min_tot)))
                            
                    use_p_src = 0
                    slew_min = 1
                    for p_src in range(1, peak_group.npksrc + 1):
                        if sr_ok[p_src - 1]:
                            if (0 < slew_max[p_src - 1]  < slew_min):
                                slew_min = slew_max[p_src - 1]
                                use_p_src = p_src

                    if use_p_src != 0:
                        p_src = use_p_src
                        k_src = peak_group.pksrnum[p_src - 1]

                        add2 = try_group2[group_index - 1] and sr_ok2[p_src - 1]
                        
                        j_scan = k_scan
                        ok_scan = 0
                        stop_k = scans[m_scan - 1].startj - slew_min
                        start_k = stop_k - peak_group.pkdwell

                        if add2:
                            n_added += 2
                            ok_scan = 2

                            stop2 = start_k - m_settle[group_index - 1] / secpday
                            start2 = stop2 - peak_group.pkdwell
                            
                            s.srinsert(k_scan, k_src, 
                                       peak_group.pksrc[p_src - 1], start2,
                                       stop2, last_scan_index, m_scan)
                            scans[k_scan - 1].point = group_index
                            
                            j_scan = k_scan + 1
                            for station_index, station in enumerate(stations):
                                if station.stascn[k_scan - 1]:
                                    last_pointing_scan[station_index] = j_scan

                        else:
                            j_scan = k_scan
                            n_added += 1
                            ok_scan = 1

                        s.srinsert(j_scan, k_src, peak_group.pksrc[p_src - 1],
                                   start_k, stop_k, last_pointing_scan, m_scan)
                        scans[j_scan - 1].point = group_index
                        
                        if s.schcon.pkwatch:
                            s.wlog(0, "ADDPEAK: Adding {} scan(s) on {} for "
                                   "pointing group {}".format(
                                       ok_scan, peak_group.pksrc[p_src - 1],
                                       group_index))
                        
                        for station_index, station in enumerate(stations):
                            if station.stascn[j_scan - 1]:
                                last_target_scan[station_index] = j_scan

                    else:
                        if s.schcon.pkwatch:
                            s.wlog(0, "ADDPEAK: No pointing sources available "
                                   "for pointing group {}".format(group_index))

                else:
                    if s.schcon.pkwatch:
                        s.wlog(0, "ADDPEAK:  No stations for group {} in this "
                               "scan. ".format(group_index))

            # end of if try_group[group_index]
        # end of loop of peak groups

        k_scan = scan_index + n_added
        if k_scan != m_scan:
            scndup(k_scan - 1, m_scan - 1, True, "ADDPEAK2")
        scans[k_scan - 1].origen = keep_origen

        n_good = s.scngeo(last_target_scan, k_scan)
        scans[k_scan - 1].point = -999

        if n_added > 0:
            peak_opt = n_added
        else:
            peak_opt = 0

    else: # peak_opt != 0
        if peak_opt > 0:
            peak_opt -= 1

    if peak_opt >= 1:
        insert_adjust = False

    return peak_opt, insert_adjust



## A python parser for SCHED's "keyin" files.
## 
## Syntax:
## !.*$ defines comments.
## KEY "=" VAL
## KEY VAL "the equals sign is optional so long as its absence doesn't lead to ambiguity"

## (1) Keys no longer than 8 chars(!)
## (1a) Except when they are ("overwrite" in first line of SCHED key files).
## (2) Keys are _not_ case sensitive(!) (normalise to GREAT RUNES in honour of FORTRAN
##     Alphabet for keywords?  [A-Za-z]*
## (3) Values are not case sensitive, except when they are. (E.g., unix paths.)
## (4) Values may be comma-separated lists ("arrays")
## (5) Values may be numbers or "character string".
## (6) Values are doubles.
## (7) Values may be expressions. (E.g., "5.5*4")
## (8) hh:mm:ss.ss dd:mm:ss.ss (no embedded blanks or signs; "2:40" is allowed and equal to "160")
## (9) Value may be blank ("").
## Entries terminated by "/"

from .util import get_catalog_dir

import schedlib

import re, sys, logging, functools
import ast
import operator as op
import itertools
import copy
import collections
import os
import json

class TokeniseError(RuntimeError): pass
class ParseError(RuntimeError): pass
class InputError(RuntimeError): pass
class EOFBeforeFirstItem(BaseException): pass

Token = collections.namedtuple("Token", ["type_", "value", "file_", "line"])

# code to support arithmatic expressions, copied from:
# https://stackoverflow.com/questions/2371436/evaluating-a-mathematical-expression-in-a-string 
      
# supported operators
operators = {ast.Add: op.add, 
             ast.Sub: op.sub, 
             ast.Mult: op.mul,
             ast.Div: op.truediv, 
             ast.Pow: op.pow, 
             ast.USub: op.neg}

def eval_expr(expr):
    return eval_(ast.parse(expr, mode="eval").body)

def eval_(node):
    if isinstance(node, ast.Num): # <number>
        return node.n
    elif isinstance(node, ast.BinOp): # <left> <operator> <right>
        return operators[type(node.op)](eval_(node.left), eval_(node.right))
    elif isinstance(node, ast.UnaryOp): # <operator> <operand> e.g., -1
        return operators[type(node.op)](eval_(node.operand))
    else:
        raise TypeError(node)

def s_expression(scanner, token):
    try:
        return ("number", float(eval_expr(token[1:-1])))
    except:
        return ("value", token)

def s_keyword(scanner, token):
    if (len(token) > 9) or ("." in token):
        res = ("value", token)
    else:
        res = ("key", token)
    return res

def s_quote(scanner, token):
    return ("quote", token[1:-1])

def s_number(scanner, token):
    return ("number", float(token))

def s_angle(scanner, token): # also time
    if token.startswith("-"):
        multiplier = -1
        l = token[1:].split(":")
    else:
        multiplier = 1
        l = token.split(":")
    ## It's not python to use reduce.
    ## But I neither remember nor care what is.
    val = functools.reduce(lambda acc, x: 60*acc+float(x), l, 0.0) 
    return ("number", multiplier * val)

def s_value(scanner, token):
    return ("value", token)

def s_string(scanner, token):
    return ("value", token)

def s_misc(scanner, token):
    return ("misc", token)

class Parser:
    """
    Parses a keyin file like object, 
    iterate_keyfile method returns the iterator
    """
    def __init__(self, input_, record_defaults={}, state_defaults={}):
        # (?=...): look-ahead, but doesn't consume
        token_separator = "(?=[\s,/]|$)" 
        self.scanner = re.Scanner([
            ("\![^\n]*", None),
            ("[ \t]+", None),
            ("\n|(\r\n)", lambda s, t: ("newline", None)),
            ("=", lambda s, t: ("equal", None)),
            ("('[^'\n]*')|(\"[^\"\n]*\")", s_quote), 
            (",", lambda s, t: ("comma", None)),
            ("[+-]?[0-9]+:[0-9]+:[0-9]+(\.[0-9]*)?" + token_separator, s_angle),
            ("[+-]?[0-9]+:[0-9]+(\.[0-9]*)?" + token_separator, s_angle),
            ("[+-]?[0-9]*\.[0-9]+(E[+-][0-9]{1,3})?(?![A-Za-z_0-9()])" + 
             token_separator, s_number),
            ("[+-]?[0-9]+\.?(?![A-Za-z_0-9()])" + token_separator, s_number),
            ("\([\d*+\-/.()E\s]+\)", s_expression),
            # keywords can be sliced
            ("[A-Za-z0-9][:()A-Za-z_0-9_+-]*(?=([ \t=/\n]|(\r\n)))", 
             s_keyword),
            ("[@\.$:()/A-Za-z_0-9\_+*-]+", s_value),
            ("[^\n]*", s_misc)
        ])
        self.set_defaults(record_defaults, state_defaults)
        self.input_stack = collections.OrderedDict()
        self.line_number = 0
        self.input_ = input_

    def _push_input(self, filename):
        self.input_stack[self.input_] = self.line_number
        self.input_ = open(filename, "r")

    def _pop_input(self):
        if len(self.input_stack) > 0:
            self.input_.close()
            self.input_, self.line_numer = self.input_stack.popitem()
        else:
            self.input_ = None

    def readline(self):
        self.line_number += 1
        if (self.input_ is sys.stdin) and self.input_.isatty():
            try:
                return input("* ") + "\n"
            except EOFError:
                return ""
        return self.input_.readline()

    def set_defaults(self, record, state):
        self.record_defaults = record
        self.state_defaults = state

    def p_skip_newline(self):
        while self.tok.type_ == "newline":
            self.tok = next(self.tokIt)

    def p_chunk(self):
        entries = []
        self.p_skip_newline()
        if self.tok.type_ == "EOF":
            raise EOFBeforeFirstItem()
        # create items until the token iterator is on a value starting with a /
        while not ((self.tok.type_ == "value") and 
                   self.tok.value.startswith("/")):
            entries.append(self.p_item())
            self.p_skip_newline()
        # read until a newline or EOF
        while self.tok.type_ not in ("newline", "EOF"):
            self.tok = next(self.tokIt)
        logging.debug("p_chunk %s", str(self.tok))
        if len(entries) == 0:
            schedlib.wlog(1, "Warning: empty record in {} line {}.".format(
                self.tok.file_.name, self.tok.line))
        return dict(entries)

    def p_item(self):
        lhs = self.p_key()
        if self.tok.type_ == "equal":
            self.tok = next(self.tokIt)
            logging.debug("p_item %s", str(self.tok))
            rhs = self.p_rhs()
        elif self.tok.type_ in ["quote", "number"]:
            rhs = self.p_rhs()
        else:
            rhs = 0.0 # for unitary expressions.
        return (lhs, rhs)

    def p_key(self):
        logging.debug("p_key: %s", str(self.tok))
        self.p_skip_newline()
        if self.tok.type_ != "key":
            if self.tok.type_ == "EOF":
                raise ParseError("You may have unused parameters at the end of "
                                 "the file (not triggered by a ‘/‘). Please "
                                 "remove them or add a '/'.")
            raise ParseError("Expected key token, got %s" % str(self.tok))

        keyword = self.tok.value.upper()
        
        self.tok = next(self.tokIt)
        return keyword

    def p_rhs(self):
        val = self.p_value()
        rhs = [val]
        while self.tok.type_ == "comma":
            logging.debug("p_rhs: %s", str(self.tok))
            self.tok = next(self.tokIt)
            rhs.append(self.p_value()) # p_value advances tok beyond the value.
        if len(rhs)==1:
            rhs = rhs[0]
        return rhs

    def p_value(self):
        self.p_skip_newline()
        if self.tok.type_ not in ["value", "quote", "number", "key"]:
            raise ParseError("Unexpected RHS token %s" % str(self.tok))
        val = self.tok
        logging.debug("p_value: %s", str(val))
        self.tok = next(self.tokIt)
        return val[1]

    def _write_defaults(self, parameters, destination):
        combined = copy.copy(self.record_defaults)
        combined.update(self.state_defaults)
        print("\n".join("{} = {}".format(
            key.upper(), 
            ", ".join(repr(e) for e in combined[key][0])
            if type(combined[key][0]) == list else
            repr(combined[key][0]))
                        for key in sorted(combined.keys())
                        if key in parameters),
              file=destination)

    # SCHED keyin backwards compatibility:
    # control keywords, they mainly support interactive usage,
    # so need to be responsive (handle it when enter is pressed, 
    # instead of waiting for a '/' end marking)
    control_re = re.compile("^\s*(" +
                            "|".join(("SAVE\s(?P<save>.*)", 
                                      "HELP\s(?P<help>.*)", 
                                      "SHOW\s(?P<show>.*)", 
                                      "@(?P<include>.*)")) +
                            ")", 
                            flags=re.IGNORECASE)
    default_file = "sched.par"
    def _handle_control(self, match):
        groups = match.groupdict()
        if groups["save"] is not None:
            destination = groups["save"].strip()
            if destination == "":
                destination = default_file
            to_save = set(self.record_defaults.keys())
            to_save.update(self.state_defaults.keys())
            with open(destination, "w") as f:
                self._write_defaults(to_save, f)
        elif groups["help"] is not None:
            print(", ".join(key.upper() for key in sorted(itertools.chain(
                self.record_defaults.keys(), self.state_defaults.keys()))))
        elif groups["show"] is not None:
            to_show = {word.lower() 
                       for word in groups["show"].split()}
            if len(to_show) == 0:
                to_show = set(self.record_defaults.keys())
                to_show.update(self.state_defaults.keys())
            self._write_defaults(to_show, sys.stdout)
        elif groups["include"] is not None:
            self._push_input(groups["include"].strip())

    def iterate_keyfile(self):
        # method intended for interactive use of KEYIN records
        tokens = []
        while True:
            # try to parse records from current tokens
            self.tokIt = iter(tokens)
            while True:
                try:
                    self.tok = next(self.tokIt)
                    yield self.p_chunk()
                except ParseError as txt:
                    raise ParseError("{} line {}: {}".format(
                        self.tok.file_.name, self.tok.line, txt))
                except EOFBeforeFirstItem:
                    # continue with next file
                    pass
                except StopIteration:
                    break
                # remove used tokens from the collection
                tokens = list(self.tokIt)

            # read more input
            if self.input_ is None:
                if len(tokens) > 0:
                    raise ParseError("Unexpected end of input")
                return

            # read line by line to allow special control sequences to be 
            # interactive, this only works if tokens are on a single line
            # (assumed to be the case)
            text = self.readline()
            if len(text) == 0:
                tokens.append(Token("EOF", "", self.input_, self.line_number))
                self._pop_input()
                continue
            
            match = self.control_re.match(text)
            if match:
                self._handle_control(match)
            else:
                res = self.scanner.scan(text)
                if res[1] != "":
                    raise ParseError("{} line {}: Unparsed text: {}.".format(
                        self.input_.name, self.line_number, res[1][:20]))
                for token in res[0]:
                    # the scanner returns the type and text
                    # add the file and line number
                    tokens.append(Token(token[0], token[1], 
                                        self.input_, self.line_number))

    def list_keyfile(self):
        # will return all KEYIN records available on the input
        tokens = []
        while True:
            text = self.readline()
            if text == "":
                tokens.append(Token("EOF", "", self.input_, self.line_number))
                self._pop_input()
                if self.input_ is None:
                    break
                else:
                    continue
            match = self.control_re.match(text)
            if match:
                self._handle_control(match)
            else:
                res = self.scanner.scan(text)
                if res[1] != "":
                    raise ParseError("{} line {}: Unparsed text: {}.".format(
                        self.input_.name, self.line_number, res[1][:20]))
                for token in res[0]:
                    # the scanner returns the type and text
                    # add the file and line number
                    tokens.append(Token(token[0], token[1], 
                                        self.input_, self.line_number))
        
        # try to parse records from current tokens
        chunks = []
        self.tokIt = iter(tokens)
        while True:
            try:
                self.tok = next(self.tokIt)
                chunks.append(self.p_chunk())
            except ParseError as txt:
                raise ParseError("{} line {}: {}".format(
                    self.tok.file_.name, self.tok.line, txt))
            except EOFBeforeFirstItem:
                left_over = list(self.tokIt)
                if len(left_over) > 0:
                    raise InputError("Unexpected end of input")
                break
            except StopIteration:
                break

        return chunks


def print_tree(res):
    logging.debug("Result: %s", str(res))
    all_chunks_text = []
    for chunk in res:
        chunk_text = []
        for (k, v) in chunk:
            logging.debug("%s %s", str(k), str(v))
            e_text = "\t%s : %s" % (repr(k), repr(v))
            j_text = re.sub("'", '"', e_text)
            chunk_text.append(j_text)
        logging.debug("Rendered: %s", chunk_text)
        all_chunks_text.append("\t{\n%s\n\t}" % (",\n".join(chunk_text)))
    logging.debug("All: %s", all_chunks_text)
    contents = ",\n".join(all_chunks_text)
    logging.debug("Contents: %s", contents)
    print("[\n%s\n]\n" % contents)

class KeyfileIterator:
    def __init__(self, input_, record_defaults={}, state_defaults={}):
        # make parser available as an ugly hack to be able to access the input
        # as reading the coverletter text is done outside of the keyin parser
        # as it is not in keyin form
        self.input_ = input_
        self.parser = Parser(input_, record_defaults, state_defaults)
        self.iterator = self.parser.iterate_keyfile()

    def __iter__(self):
        return self.iterator

    def __next__(self):
        return next(self.iterator)

    def set_defaults(self, record, state):
        """
        This is used to show interactive help
        """
        self.parser.set_defaults(record, state)

class KeyfileLister:
    def __init__(self, input_, record_defaults={}, state_defaults={}):
        # make parser available as an ugly hack to be able to access the input
        # as reading the coverletter text is done outside of the keyin parser
        # as it is not in keyin form
        self.input_ = input_
        self.parser = Parser(input_, record_defaults, state_defaults)
        self.iterator = None

    def init_iterator(self):
        if self.iterator is None:
            # if the input is a file in the SCHED catalogs directory, 
            # check cache
            input_split = os.path.split(self.input_.name)
            catalog_dir = get_catalog_dir()
            try:
                is_catalog_dir = os.path.samefile(input_split[0], catalog_dir)
            except:
                is_catalog_dir = False

            if is_catalog_dir:
                cache_filename = os.path.join(catalog_dir, "cache",
                                              input_split[1])
                try:
                    cache_up_to_date = os.path.getmtime(self.input_.name) < \
                                       os.path.getmtime(cache_filename)
                except:
                    cache_up_to_date = False
                if cache_up_to_date:
                    with open(cache_filename, "r") as cache_file:
                        data = json.load(cache_file)
                else:
                    data = self.parser.list_keyfile()
                    try:
                        with open(cache_filename, "w") as cache_file:
                            json.dump(data, cache_file)
                    except:
                        pass
                self.iterator = iter(data)
            else:
                self.iterator = iter(self.parser.list_keyfile())

    def __iter__(self):
        self.init_iterator()
        return self.iterator

    def __next__(self):
        self.init_iterator()
        return next(self.iterator)

    def set_defaults(self, record, state):
        """
        This is used to show interactive help
        """
        assert(self.iterator is None)
        self.parser.set_defaults(record, state)

def iterate_keyfile(f):
    return Parser(f).iterate_keyfile()

def read_keyfile(f):
    return Parser(f).list_keyfile()

def process_file(f):
    res = read_keyfile(f)
    print_tree(res)

def convert(fnin, fnout):
    with open(fnin, "r") as i:
        j = read_keyfile(i)
    with open(fnout, "w") as o:
        json.dump(j, o)
    
if __name__=="__main__":
    import sys, logging
    logging.basicConfig(filename="example.log", level=logging.DEBUG)
    logging.debug("Srsly you guys")
    convert(sys.argv[1], sys.argv[2])
    



from ..catalog import FrequencyCatalog
from .. import util
from ..key import read_keyfile

import schedlib as s

def getfreq():
    freqfile = bytes(s.schsco.freqfile).decode().strip()
    
    s.wlog(0, "GETFREQ: Reading frequency file:  {}".format(freqfile))
    
    try:
        f = open(freqfile, "r")
    except Exception as e:
        s.wlog(1, str(e))
        s.errlog("GETFREQ: Problem opening frequency file")
    else:
        with f:
            keyin_data = read_keyfile(f)
    
    catalog = FrequencyCatalog()
    
    if len(keyin_data) > catalog.maxfreq:
        s.errlog("GETFREQ: Too many frequency groups. ")

    maxsta = s.frqc.fstnam.shape[0]
    state_defaults = {
        "version": ["", util.noop]
    }
    record_defaults = {
        "station":  [[""] * maxsta, util.foreach(util.upper)],
        "name":     ["",            util.noop],
        "note":     ["",            util.noop],
        "priority": [0.,            util.noop],
        "ifname":   [[],            util.noop],
        "altifn":   [[],            util.noop],
        "fe":       [[],            util.foreach(util.lower)],
        "pol":      [[],            util.foreach(util.upper)],
        "rf1":      [[],            util.noop],
        "rf2":      [[],            util.noop],
        "lo1":      [[],            util.noop],
        "ch1rf1":   [[],            util.noop],
        "ch1rf2":   [[],            util.noop],
        "dualx":    [1.,            util.to_bool],
        "syn":      [[0., 0., 0.],  util.noop],
        "lcp50cm":  ["NARROW",      util.upper],
        "rcp50cm":  ["NARROW",      util.upper],
        "vlabw":    [None,          util.noop], # not used
        "vlaband":  [None,          util.noop], # not used
        "flukea":   [None,          util.noop], # not used
        "flukeb":   [None,          util.noop], # not used
        "vlafeab":  [None,          util.noop], # not used
        "vlafecd":  [None,          util.noop], # not used
        "vlasyna":  [None,          util.noop], # not used
        "vlasynb":  [None,          util.noop], # not used
        "fefilter": [None,          util.noop], # not used
        "tscal":    [None,          lambda x: x if x is None else x.upper()]
    }

    # map from frequency catalog entry attributes to keyin keywords
    attribute_to_key = {
        "frname": "name",
        "frnote": "note",
        "fstnam": "station",
        "prio": "priority",
        "fifnam": "ifname",
        "faltif": "altifn",
        "tscal": "tscal"
    }
    # above are the exceptions to the rule: remove first 'f' to go from 
    # attribute to key
    attribute_to_key.update({
        attribute: attribute[1:] for attribute in catalog.attributes
        if attribute not in attribute_to_key.keys()})

    for index, record in enumerate(keyin_data):
        values, present = util.merge_record_with_defaults(
            record, record_defaults, state_defaults)

        # copy all values to the common block placeholder
        entry = catalog.entries[index]
        entry.set_keyin_values(values, attribute_to_key)
        if entry.tscal not in (None, "CONT", "GAP"):
            raise RuntimeError("Unknown TSCAL in frequency catalog for {} "
                               "entry {}:  {}".format(
                                   entry.station, entry.name, entry.tscal))
    
    catalog.write(range(len(keyin_data)))

    s.frqn.nfreq = len(keyin_data)
    s.schsco.freqver =  util.resize_string(state_defaults["version"][0], 
                                           s.schsco.freqver.itemsize, "version")

    s.listfreq()



from . import parameter

import schedlib as s

def gettim(values, entries, index, start, stop, day, year, mjd1):
    day[index] = values["day"]
    month = values["month"]
    year[index] = values["year"]
    if year[index] < 100 and year[index] != 0:
        s.errlog("GETTIM:  Don''t use 2 digit year!")

    if (day[index] != 0) and (month > 1):
        year[index], day[index], error = s.sla_calyd(
            year[index], month, day[index])
        if error == 2:
            s.errlog("GETTIM: Bad month ")
        if error == -1:
            s.errlog("GETTIM: Unlikely year ")

    start[index] = values["start"] / 86400. \
                   if values["start"] != parameter.unset else parameter.unset
    stop[index] = values["stop"] / 86400. \
                  if values["stop"] != parameter.unset else parameter.unset

    entry = entries[index]

    if index == 0:
        if day[index] > 366:
            myear, mday, mut = s.lst2ut(0., day[index], 0.)
            mjd1, error = s.sla_cldj(myear, 1, mday)
            if error not in [0, 3]:
                s.errlog("GETTIM: Problem converting day to MJD")
        else:
            mjd1, error = s.sla_cldj(year[index], 1, day[index])
            if error not in [0, 3]:
                s.errlog("GETTIM: Problem getting MJD of first scan before "
                         "station catalog read.")

    if (values["dwell"][0] != parameter.unset) and \
       (values["duration"] != parameter.unset):
        s.errlog("GETTIM: Don''t specify both DUR and DWELL!  Input scan: {}".\
                 format(index+1))
    elif values["dwell"][0] != parameter.unset:
        entry.dwell = True
        entry.dur = values["dwell"][0] / 86400.
        entry.nowait = values["dwell"][1] 
        entry.mindw = values["dwell"][2] / 86400.
    elif values["duration"] != parameter.unset:
        entry.dwell = False
        entry.dur = values["duration"] / 86400.
        entry.nowait, entry.mindw = (0, 0.)
    else:
        if index == 0:
            entry.dwell = False
            entry.dur, entry.nowait, entry.mindw = (0., 0, 0.)
        else:
            prev = entries[index - 1]
            entry.dwell, entry.dur, entry.nowait, entry.mindw = \
                prev.dwell, prev.dur, prev.nowait, prev.mindw

    s.schn1.dwells |= entry.dwell
    return mjd1



from . import schin, getfreq, rdpeak_open, srread_open, getset, \
    parameter
from .. import util

import schedlib as s

def input_(stdin):
    if s.schcon.debug:
        s.wlog(0, "INPUT starting")
    schin(stdin)

    rdpeak_open(stdin)

    s.accsrc(True)
    s.schcsc.srver = " --- ".ljust(s.schcsc.srver.itemsize)
    s.schcsc.srver2 = " --- ".ljust(s.schcsc.srver2.itemsize)
    srread_open(util.f2str(s.schcsc.srcfile), stdin, True, '1')
    srcfile2 = util.f2str(s.schcsc.srcfile2)
    if (srcfile2 != "NONE"):
        srread_open(srcfile2, stdin, True, '3')
    if (s.schpeakn.dopoint):
        srread_open(util.f2str(s.schpeakc.psrcfile), stdin, True, '2')
    if (not s.schcon.noset):
        getset()
    if (not s.schcon.noset):
        getfreq()

    s.wlog(0, "INPUT:   Found {:5d} input scans.".format(int(s.schn1.nscans)))



from .catalog import SetupCatalog, ScanCatalog, StationCatalog, SourceCatalog, \
                     FrequencyCatalog
from .util import f2str
from . import dbbc_patching, vex_scans
from .sched import parameter

import schedlib as s

import numpy as np

import re
from collections import defaultdict, OrderedDict
import itertools
import functools
import copy
import os.path

sched_version = "2" # version of SCHED VEX writing routine

block_separator = "*------------------------------------------------------"\
                  "------------------------\n"
def write(output, vex_version="2.0", print_warnings=False):
    assert vex_version in {"1.5", "2.0"}
    # read in all catalogs, so other methods can assume their entries are valid
    for Catalog in (SetupCatalog, ScanCatalog, StationCatalog, SourceCatalog,
                    FrequencyCatalog):
        Catalog().read()
    station_catalog = StationCatalog()
    station_catalog.read_scheduled_attributes()
    SourceCatalog().set_aliases()
    scan_catalog = ScanCatalog()
    vex_scans.apply_tape_offset(
        scan_catalog.used(), scan_catalog.scan_offset, 
        station_catalog.used(), SetupCatalog().scheduled())

    output.write(header_block(vex_version))
    output.write(block_separator)
    output.write(global_block(vex_version))
    output.write(block_separator)
    output.write(exper_block(vex_version))
    output.write(block_separator)
    mode_text, scan_mode_name = modes_block(vex_version, print_warnings)
    output.write(mode_text)
    output.write(block_separator)
    output.write(stations_block(vex_version, print_warnings))
    output.write(block_separator)
    output.write(procedures_block(vex_version))
    output.write(block_separator)
    output.write(source_block(vex_version, print_warnings))
    output.write(block_separator)
    output.write(sched_block(scan_mode_name, vex_version, print_warnings))

def header_block(vex_version):
    return """
VEX_rev = {};
*    SCHED vers: {}
*    VEX/SCHED:  {}.{}
*    Other versions: Sched: {:5.1f} Plot: {:4.2f} JPL-ephem: {:4.2f}
""".format(vex_version,
           f2str(s.verc.version),
           vex_version, sched_version,
           float(s.vern.vernum), float(s.plver()), float(s.jplver()))[1:]

def global_block(vex_version):
    return """
$GLOBAL;
     ref $EXPER = {};
*                                                      +------------+
*                         PI revision number:          |{:10.4f}  |
*                                                      +------------+
*    mode:       {}
""".format(str2def(f2str(s.schc1.expcode)),
           float(s.schco.schver),
           f2str(s.schsco.obsmode))[1:]

def exper_block(vex_version):
    expcode = f2str(s.schc1.expcode)
    ret  = """
$EXPER;
*
def {};
     exper_name = {};
     exper_description = {};
     PI_name = {};
     PI_email = {};
""".format(str2def(expcode),
           str2val(expcode),
           str2val(f2str(s.schc1.expt)),
           str2val(f2str(s.schsco.piname)),
           str2val(f2str(s.schsco.email)))[1:]

    ret += """
*    address:   {}
*               {}
*               {}
*               {}
""".format(*(f2str(address) for address in s.schsco.address))[1:]

    ret += """
*    phone:     {}
*    during obs:{}
*    fax:       {}
""".format(*(f2str(contact) for contact in [s.schsco.phone,
                                            s.schsco.obsphone,
                                            s.schsco.fax]))[1:]

    comment_prefix = "*    "
    ret += text2comments((f2str(note) for note in s.schsco.note), 
                         comment_prefix)

    catalog = ScanCatalog()
    scans = catalog.used()
    year, doy, time_ = s.timej(scans[0].stopj)
    year, month, day, jd, mname, dname = s.tdatecw(year, 1, doy)
    ret += """
*
*    year, doy: {}, {}
*    date     : {} {} {} {}
*    MJD      : {}
""".format(year, doy,
           f2str(dname), day, f2str(mname), year,
           int(jd - 2400000.5 ))[1:]

    max_stations = s.schn1.stanum.shape[0]
    ret += "     exper_nominal_start={};\n".format(time2str(scans[0].startj))
    ret += "     exper_nominal_stop={};\n".format(time2str(scans[-1].stopj))

    if s.schcon.coverlet:
        with open("schedcov.tmp", "r") as f:
            ret += "*\n"
            ret += "*-------------------- cover  letter --------------------\n"
            for line in f.readlines():
                ret += "*" + line
            ret += "*------------------------- end -------------------------\n"


    if s.schsco.correl:
        ret += """
*
     target_correlator = {};
*
""".format(str2val(f2str(s.schsco.correl)))[1:]

    for text, value in (
            ("*    integr_time    : {} s", float(s.schco.coravg)),
            ("*    number_channels: {}",   s.schco.corchan),
            ("*    number_antenna : {}",   s.schco.cornant),
            ("*    cross_polarize : {}",   "Yes" if s.schco.corpol else "No"),
            ("*    weight_func    : {}",   f2str(s.schsco.corwtfn)),
            ("*    distrib_medium : {}",   f2str(s.schsco.cortape)),
            ("*    distrib_format : {}",   f2str(s.schsco.cordfmt)),
            ("*    source_pos_cat : {}",   f2str(s.schsco.corsrcs))):
        if value:
            ret += text.format(value) + "\n"

    if s.schsco.corship[0]:
        ret += "*    distribute_to  :\n"
        for ship in s.schsco.corship:
            if ship:
                ret += "*                     {}\n".format(f2str(ship))

    
    ret += text2comments((f2str(note) for note in s.schsco.cornote), 
                         comment_prefix)
    ret += """
*
enddef;
"""[1:]

    return ret

def modes_block(vex_version, print_warnings):
    # some common block variables are indexed by scan number, but not part 
    # of the scan catalog, the indices are offset from the catalog entries
    scan_catalog = ScanCatalog()
    scans = scan_catalog.used()
    scan_offset = scan_catalog.scan_offset
    
    setups = SetupCatalog().scheduled()

    stations = StationCatalog().used()

    format_map = {"MARKIII": "Mark3A",
                  "MKIV": "Mark4",
                  "VLBA": "VLBA",
                  # 2 valid LBA formats in vex2 LBA_AT and LBA_VSOP, 
                  # no idea which is supposed to be used.
                  # For VEX1 only the fact that it starts with LBA is used.
                  "LBA": "LBA_AT",
                  "S2": "S2",
                  "MARK5B": "MARK5B",
                  "VDIF": "VDIF"} 
                  
    phase_cals = defaultdict(lambda: defaultdict(set))
    pcal_map = {"OFF" : 0,
                "1MHZ": 1,
                "5MHZ": 5,
                ""    : 1}
    def cmp_bitstream(x, y):
        (sideband_a, bbc_a, channel_a) = x
        (sideband_b, bbc_b, channel_b) = y
        if sideband_a != sideband_b:
            return 1 if sideband_a == "L" else -1
        if bbc_a != bbc_b:
            return bbc_a - bbc_b
        return channel_a - channel_b

    def do_if(setup, if_pcal):
        # with duplicate IFs, use the first (mimic SCHED)
        if_channel = OrderedDict()
        for channel, if_ in enumerate(setup.ifchan):
            if if_ not in if_channel:
                if_channel[if_] = channel
        
        if_ = tuple(("if_def",
                     "&IF_" + if_name,
                     if_name if vex_version < "2" else "",
                     setup.pol[channel][0], # only take first character (RCP->R)
                     "{:9.2f} MHz".format(setup.firstlo[channel]),
                     setup.side1[channel]) +
                    tuple("{:d} MHz".format(i) for i in if_pcal)
                    for if_name, channel in if_channel.items())
        return if_, if_channel
        
    def do_bbc(setup, if_, if_channel):
        # BBC numbers brought into range 1-16 for drudg
        bbc_channel = OrderedDict(
            [((bbc - 1) % 16 + 1, channel) 
             for channel, bbc in enumerate(setup.bbc)])
        bbc = tuple(("BBC_assign",
                     "&BBC{:02d}".format(bbc_number),
                     bbc_number,
                     # get IF link name
                     if_[tuple(if_channel).index(setup.ifchan[channel])][1])
                    for bbc_number, channel in bbc_channel.items())
        return bbc, bbc_channel

    def do_phasecal_and_freq(setup, scan, tone_interval, bbc, bbc_channel):
        # get LO sum and BBC width from schedule if specified,
        # else from setup
        lo_sum = [(scan.freq[channel_index] + 
                   setup.corinv[channel_index])
                  if scan.freq[channel_index] > 0 else
                  (setup.firstlo[channel_index] + 
                   (1 if setup.side1[channel_index] == "U" else -1) * 
                   setup.bbsyn[channel_index])
                  for channel_index in range(setup.nchan)]
        bbc_width = [scan.bw[channel_index]
                     if scan.bw[channel_index] > 0 else
                     setup.bbfilt[channel_index]
                     for channel_index in range(setup.nchan)]
        
        if tone_interval != 0:
            ntone, tone_channel, ntone_per_def, itone_per_def = \
                s.vxton2(-1, 
                         scan.setnum, 
                         lo_sum, 
                         setup.netside, 
                         bbc_width, 
                         tone_interval)
            if ntone > 1:
                link = "&PCD{}"
            else:
                link = "&PCD"
            phase_cal = tuple(("phase_cal_detect", link.format(phase_def + 1)) +
                              tuple(itone_per_def[:ntone_per_def[phase_def], 
                                                  phase_def])
                              for phase_def in range(ntone))
        else:
            # ``empty'' phase cal and tone channel map
            phase_cal = (("phase_cal_detect", "&NoCal"),)
            tone_channel = [1 for channel_index in range(setup.nchan)]
        
        freq = tuple(("chan_def",
                      "",
                      "{:.6f}".format(lo_sum[channel_index]).rstrip("0").\
                      rstrip(".") + " MHz",
                      setup.netside[channel_index],
                      "{:7.2f} MHz".format(bbc_width[channel_index])
                      if bbc_width[channel_index] >= 1 else
                      "{:7.3f} kHz".format(bbc_width[channel_index] * 1000),
                      "&CH{:02d}".format(channel_index + 1),
                      # BBC link name
                      bbc[tuple(bbc_channel).index(
                          setup.bbc[channel_index])][1],
                      # phase cal link name, tone_channel is fortran
                      # based, so -1 for array indexing
                      phase_cal[tone_channel[channel_index] - 1][1])
                     for channel_index in range(setup.nchan))
        if vex_version < "2":
            freq += (("sample_rate", "{:7.3f} Ms/sec".format(setup.samprate)),)
        return phase_cal, freq

    def do_track(setup, station, track_format, bbc=None): 
        # bbc represents the $BBC VEX section, only used for S2 format in VEX1
        
        if track_format is None:
            track_format = "NONE"
        
        fanout = int(setup.fanout)
        if track_format.startswith("LBA") or track_format.startswith("MARK5B"):
            def track_number(base_track, bit_index, fanout_index):
                return base_track + bit_index
        elif track_format.startswith("VDIF"):
            def track_number(base_track, bit_index, fanout_index):
                return base_track
        else:
            def track_number(base_track, bit_index, fanout_index):
                return base_track + fanout_index * 2 + \
                    bit_index * 2 * fanout

        if vex_version < "2":
            if track_format == "NONE":
                return (("track_frame_format", "NONE"),)
            elif track_format.startswith("S2"):
                iset = setups.index(setup) + 1
                s2mdnm, valid, ns2usd, is2usd = s.vxs2md(iset, True)
                if not valid:
                    raise RuntimeError("Inconsistent or impossible S2 mode")
                track = (("S2_recording_mode", f2str(s2mdnm)),)
                if station.dar == "MKIV":
                    # link to BBC 1 and 2
                    track += (("S2_data_source", bbc[0][1], bbc[1][1]),)
                elif station.dar in {"VLBA", "VLBAG"}:
                    bbc_numbers = {c.bbc for c in setup.channel}
                    if bbc_numbers <= {1, 2, 3, 4}:
                        track += (("S2_data_source", "VLBA_BBC_1-4"),)
                    elif bbc_numbers <= {5, 6, 7, 8}:
                        track += (("S2_data_source", "VLBA_BBC_5-8"),)
                    else:
                        raise RuntimeError("Inconsistent DAR connection")
                elif station.dar == "NONE":
                    track += (("S2_data_source", "none"), )
                else:
                    track += (("S2_data_source", station.dar), )
                # no fanout_defs
                return track
                
            elif track_format.startswith("LBA"):
                track = (("S2_data_source", "VLBA"),
                         ("S2_recording_mode", "none"))
            elif track_format.startswith("VDIF"):
                # 5032 functionality copied from SCHED, undocumented VEX,
                # not used by JIVE
                track = (("track_frame_format", "VDIF5032"),)                
            else:
                track = (("track_frame_format", track_format),)
                if track_format == "Mark3A":
                    track += (("data_modulation", "off"),)
                if track_format == "VLBA":
                    track += (("data_modulation", "on"),)

        else: # vex_version >= "2"
            track = (("track_frame_format", track_format),)
            track += (("sample_rate", "{:7.3f} Ms/sec".format(setup.samprate)),)
            
        for channel_index in range(setup.nchan):
            base_track_number = setup.track[0, channel_index]
            if s.schn5.twohead and (base_track_number > 33):
                base_track_number -= 64
                headstack = 2
            else:
                headstack = 1

            track += tuple(itertools.chain(
                [("fanout_def",
                  "",
                  freq[channel_index][5], # FREQ link name
                  bit,
                  headstack) +
                  tuple(track_number(base_track_number, bit_index, 
                                 fanout_index) 
                        for fanout_index in range(fanout))
                 for bit_index, bit in enumerate(["sign", "mag"][
                         :int(setup.bits[channel_index])])]))
        return track
        
    def do_datastream_type(setup, freq, station, track_format): 
        assert(vex_version >= "2")
        if track_format is None:
            return None, None, None
        
        def datastream_for_channels(datastream_link, channel_links):
            thread_link = "&thread0"
            datastream = (("datastream", datastream_link, "VDIF"),
                          ("thread", 
                           datastream_link, 
                           thread_link, 
                           0, # thread number
                           len(channel_links),
                           "{:7.3f} Ms/sec".format(setup.samprate),
                           setup.bits[0],
                           "real",
                           8000)) + \
                tuple(("channel",
                       datastream_link,
                       thread_link,
                       channel_link,
                       channel_in_thread)
                       for channel_in_thread, channel_link in enumerate(
                               channel_links))
            return datastream

        if station.dar == "DBBC3":
            if not track_format.startswith("VDIF"):
                raise RuntimeError(
                    "SCHED can only create VDIF VEX files for the DBBC3, "
                    "data format {} is not supported for {}.".format(
                        track_format, station.station))
            # create one data stream per IF
            used_ifs = {channel.ifchan for channel in setup.channel}
            # sort by frequency and sideband, L(ower) before U(pper)
            channel_order = sorted(setup.channel, 
                                   key=lambda x: (x.freqref, x.sidebd))
            
            datastream = ()
            for index, if_ in enumerate(sorted(used_ifs), 1):
                datastream_link = "&DS{}".format(index)
                channel_links = [
                    freq[index][5] # FREQ link name
                    for index, channel in enumerate(channel_order)
                    if channel.ifchan == if_]
                datastream += datastream_for_channels(datastream_link, 
                                                      channel_links)
            return None, datastream, None


        channel_order = sorted(
            ((setup.sidebd[channel_index],
              setup.bbc[channel_index],
              channel_index)
             for channel_index in range(setup.nchan)),
            key=functools.cmp_to_key(cmp_bitstream))
        channel_order = ((sideband,
                          bbc,
                          bit,
                          channel_index)
                         for sideband, bbc, channel_index 
                         in channel_order
                         for bit in ["sign", "mag"][
                                 :int(setup.bits[channel_index])])
                
        if track_format.startswith("VDIF"):
            # hard-coded 1 thread VDIF section, with astro patching
            datastream_link = "&DS1"
            channel_links = [freq[channel[3]][5] for channel in channel_order
                             if channel[2] == "sign"]
            datastream = datastream_for_channels(datastream_link, 
                                                 channel_links)
                                                 
            return None, datastream, None

        elif track_format.startswith("MARK5B"):
            try:
                channel_data = [(setup.sidebd[channel_index],
                                 setup.bbc[channel_index],
                                 bit)
                                for channel_index in range(setup.nchan)
                                for bit in ["sign", "mag"][
                                        :int(setup.bits[channel_index])]]
                input_bitstreams = dbbc_patching.get_input_bitstreams(
                    channel_data)
                # on-disk bitstreams are the compressed input bitstreams
                sorted_bitstreams = sorted(input_bitstreams)
                on_disk_bitstreams = (sorted_bitstreams.index(i)
                                      for i in input_bitstreams)
                bitstream_data = zip(input_bitstreams, 
                                     on_disk_bitstreams,
                                     (d[2] for d in channel_data), # sign/mag
                                     range(setup.nchan))
            except dbbc_patching.NoMatchingPatch:
                bitstream_data = ((bitstream, bitstream, bit, channel_index)
                                  for bitstream, (_, _, bit, channel_index) 
                                  in enumerate(channel_order))
            bitstream = (("stream_sample_rate",
                          "{:7.3f} Ms/sec".format(setup.samprate)),) + \
                        tuple(("stream_def",
                               freq[channel_index][5], # FREQ link name
                               bit,
                               input_bitstream,
                               on_disk_bitstream)
                              for (input_bitstream, on_disk_bitstream, 
                                   bit, channel_index) in bitstream_data)
            return bitstream, None, None

        else:
            track = do_track(setup, station, track_format, vex_version)
            return None, None, track

    def do_roll(setup):
        roll_by = -1
        if setup.barrel == "roll_auto":
            if setup.format.startswith("MARKIII"):
                roll_by = 0
            else:
                if setup.tapemode in {1, 2}:
                    roll_by = 16
                elif setup.tapemode in {4, 8}:
                    roll_by = 8
                elif setup.tapemode == 0:
                    roll_by = 0
        else:
            roll_by = {"roll_off": 0,
                       "roll_16": 16,
                       "roll_8": 8}.get(setup.barrel, -1)

        if roll_by == -1:
            raise RuntimeError("Unknown barrel-roll: {}".format(setup.barrel))

        if roll_by == 0:
            return (("roll", "off"),)
        else:
            roll = (("roll", "on"),
                    ("roll_reinit_period", "2 sec"),
                    ("roll_inc_period", 1))

        heads = 1
        if s.schn5.twohead and any(max(channel.track[:setup.tapemode-1]) >= 64
                                   for channel in setup.channel):
            heads = 2

        n_groups = 4
        for head in range(1, heads + 1):
            if roll_by == 8:
                for group in range(n_groups):
                    if group <= n_groups / 2:
                        min_track = group * 2 * 32 / n_groups + 2
                    else:
                        min_track = (group - n_groups / 2) * 2 * 32 / n_groups \
                                    + 2
                    for i in range(1, 32 / n_groups + 1):
                        if group < n_groups / 2:
                            track = group * 2 * 32 / n_group + i * 2
                        else:
                            track = (group - n_groups / 2) * 2 * 32 / n_groups \
                                    + i * 2 + 1

                        roll_def = ("roll_def", head, track)
                        for j in range(8):
                            roll_to = track - 2 * j
                            if roll_to < min_track:
                                roll_to += 2 * 32 / n_groups
                            roll_def += (roll_to,)
                        roll += (roll_def,)
            elif roll_by == 16:
                for group in range(n_groups):
                    for i in range(1, 32 / n_groups + 1):
                        if group < n_groups / 2:
                            track = group * 2 * 32 / n_groups + i * 2
                        else:
                            track = (group - n_groups / 2) * 2 * 32 / n_groups \
                                    + i * 2 + 1

                        roll_def = ("roll_def", head, track)
                        for j in range(16):
                            if j < 8:
                                alt_group = group
                                j_alt = j
                            else:
                                if group % 2 == 1:
                                    alt_group = group - 1
                                else:
                                    alt_group = group + 1
                                j_alt = j - 8
                                
                            if alt_group < n_groups / 2:
                                roll_to = alt_group * 2 * 32 / n_groups + \
                                          i * 2 - j_alt * 2
                                min_track = alt_group * 2 * 32 / n_groups + 2
                            else:
                                roll_to = (alt_group - n_groups / 2) * 2 * 32 \
                                          / n_groups + i * 2 - j_alt * 2 + 1
                                min_track = (alt_group - n_groups / 2) * 2 * 32\
                                            / n_groups + 2 + 1

                            if roll_to < min_track:
                                roll_to = roll_to + 2 * 32 / n_groups

                            roll_def += (roll_to, )
                        
                        roll += (roll_def,)

    def do_pass_order(setup, track_format):
        if track_format is None:
            return None

        if track_format.startswith("LBA"):
            return (("S2_group_order", 0),)

        if track_format == "S2":
            group_order = ("S2_group_order",) + tuple(range(setup.tapemode))
            return (group_order,)
            
        return None
    
    blocks = ["IF", "BBC", "PHASE_CAL_DETECT", "FREQ", "TRACKS", 
              "ROLL", "PASS_ORDER", # VEX 1
              "BITSTREAMS", "DATASTREAMS"] # VEX 2
    # mode is mappings from block type to a 
    # <mapping from block def to stations>
    modes = [{block: defaultdict(set) for block in blocks}]
    # mapping from scan index to index in modes list
    scan_mode_index = {}
    # mapping from mode index to set of setup files used
    mode_index_setups = defaultdict(set)

    def add_to_mode(mode, scan_mode):
        for block, block_defs in scan_mode.items():
            for block_def, block_stations in block_defs.items():
                mode_stations_in_other_defs = set()
                for mode_block_def, mode_stations in mode[block].items():
                    if (mode_block_def != block_def) and \
                       (block_stations & mode_stations):
                        # this mode has stations for the same block type
                        # in another block def, so it's not compatible
                        return False
        # everything is compatible, add the stations to the current mode
        for block, block_defs in scan_mode.items():
            for block_def, block_stations in block_defs.items():
                mode[block][block_def].update(block_stations)
        return True

    do_phase_cal_warning = False
    for scan_index, scan in enumerate(scans):
        scan_mode = {block: defaultdict(set) for block in blocks}

        setups_used = set()
        for station in stations:
            setup = setups[station.nsetup[scan_index+scan_offset] - 1]
            station_code = station.stcode
            if station.stascn[scan_index+scan_offset]:
                setups_used.add(setup)
                # add pcal to IF def if pcal was specified in the schedule
                frequency_setup_index = station.fseti[scan_index+scan_offset]
                if_pcal = ()
                if frequency_setup_index > 0:
                    tone_interval = pcal_map[
                        f2str(s.fsetc.fspcal[frequency_setup_index-1]).upper()]
                    if tone_interval != 0:
                        if_pcal = (tone_interval,)
                else:
                    tone_interval = pcal_map[setup.spcal.upper()]
                if tone_interval != 1:
                    do_phase_cal_warning = True
                
                if_, if_channel = do_if(setup, if_pcal)
                scan_mode["IF"][if_].add(station_code)

                bbc, bbc_channel = do_bbc(setup, if_, if_channel)
                scan_mode["BBC"][bbc].add(station_code)

                try:
                    phase_cal, freq = do_phasecal_and_freq(
                        setup, scan, tone_interval, bbc, bbc_channel)
                except ValueError:
                    raise RuntimeError(
                        "Inconsistent or incomplete frequency setup for {} "
                        "in setup {}.".format(station.station, setup.setname))
                scan_mode["PHASE_CAL_DETECT"][phase_cal].add(station_code)
                scan_mode["FREQ"][freq].add(station_code)
                
                track_format = next(
                    (track_format for sched_format, track_format in 
                     format_map.items()
                     if setup.format.startswith(sched_format)),
                    None)
                if vex_version < "2":
                    track = do_track(
                        setup, station, track_format, bbc)
                    scan_mode["TRACKS"][track].add(station_code)

                    roll = do_roll(setup)
                    scan_mode["ROLL"][roll].add(station_code)

                    pass_order = do_pass_order(setup, track_format)
                    if pass_order is not None:
                        scan_mode["PASS_ORDER"][pass_order].add(station_code)
                else:
                    bitstream, datastream, track = do_datastream_type(
                        setup, freq, station, track_format)
                    if bitstream is not None:
                        scan_mode["BITSTREAMS"][bitstream].add(station_code)
                    if datastream is not None:
                        scan_mode["DATASTREAMS"][datastream].add(station_code)
                    if track is not None:
                        scan_mode["TRACKS"][track].add(station_code)
        
        mode_index = next((index for index, mode in enumerate(modes) 
                           if add_to_mode(mode, scan_mode)), 
                          None)
        if mode_index is None:
            # current scan mode doesn't fit any of the current modes
            mode_index = len(modes)
            modes.append(scan_mode)
        scan_mode_index[scan_index] = mode_index
        mode_index_setups[mode_index].update(s.setname for s in setups_used)

    if do_phase_cal_warning and (not s.schn5.allvlba) and print_warnings:
        s.wlog(1, "WARNING, Phase cal not under computer control for some "
               "stations.")
        s.wlog(1, "Contact stations by e-mail to make sure phase cal is "
               "switched off.")
        s.wlog(1, "Phase cal is under computer control for VLBA stations and "
               "many others.")

    if (len(modes) > 20) and print_warnings:
        s.wlog(1, "WARNING: More than 20 VEX modes ({}) in this schedule. ".\
               format(len(modes)))
        s.wlog(1, "This VEX will NOT run in the field system!")
        s.wlog(1, "It should be ok on the VLBA.")
    
    # functions to name block defs
    def make_if_name(block_def):
        pols = set(if_def[3] for if_def in block_def)
        freq_split = block_def[0][4].split() # "4281.1 MHz" -> ["4281.1", "MHz"]
        return "LO@{}{}{}Pol{}".format(
            round(float(freq_split[0])),
            freq_split[1],
            "D" if len(pols) > 1 else pols.pop(),
            "NoTone" if len(block_def[0]) < 7 else "Tone/{}".format(
                block_def[0][6].split()[0]))

    def make_bbc_name(block_def):
        return "{}BBCs".format(len(block_def))

    def make_phasecal_name(block_def):
        return "NoDetect" if len(block_def[0]) < 3 else "Detect"
    
    def make_freq_name(block_def):
        chan_defs = [param_values for param_values in block_def
                     if param_values[0] == "chan_def"]
        sky_freq_split = chan_defs[0][2].split()
        bandwidth_split = chan_defs[0][4].split()
        return "{}{}{}x{}{}".format(
            sky_freq_split[0], sky_freq_split[1], len(chan_defs),
            round(float(bandwidth_split[0])), bandwidth_split[1])
    
    def make_datastreams_name(block_def):
        param_values = defaultdict(list)
        for d in block_def:
            param_values[d[0]].append(d)
        return "{}_{}-{}Thr{}Ch".format(
            param_values["datastream"][0][2], # VDIF/VDIF_legacy
            param_values["thread"][0][8], # bytes/packet
            len(param_values["thread"]),
            len(param_values["channel"]))

    def make_bitstreams_name(block_def):
        channels = set(param_values[1] for param_values in block_def
                       if param_values[0] == "stream_def")
        bits = set(param_values[2] for param_values in block_def
                   if param_values[0] == "stream_def")
        return "BS{}Ch{}Bit".format(len(channels), len(bits))

    def make_tracks_name(block_def):
        track_format = next((param_values[1] for param_values in block_def
                             if param_values[0] == "track_frame_format"),
                            None)
        if track_format is None:
            if any(param_values[0] == "S2_recording_mode"
                   for param_values in block_def):
                track_format = "S2"
            else:
                track_format = "NoTrackFormat"
        fanout = next((len(param_values) - 5 for param_values in block_def
                       if param_values[0] == "fanout_def"),
                      0)
        channels = set(param_values[2] for param_values in block_def
                       if param_values[0] == "fanout_def")
        bits = set(param_values[3] for param_values in block_def
                   if param_values[0] == "fanout_def")
        return "{}.{}Ch{}bit1to{}".format(track_format,
                                          len(channels),
                                          len(bits),
                                          fanout)

    def make_roll_name(block_def):
        on_off = next((param_values[1] for param_values in block_def
                       if param_values[0] == "roll"),
                      "off")
        if on_off == "on":
            roll_def = next((param_values for param_values in block_def
                             if param_values[0] == "roll_def"),
                            None)
            if roll_def is None:
                return "RollNone"
            # count the number of alternative tracks,
            # subtract "roll_def" head and track number
            return "Roll{}".format(len(roll_def) - 3)
        else:
            return "NoRoll"

    def make_pass_order_name(block_def):
        passes = next((len(param_values) - 1 for param_values in block_def
                       if param_values[0] == "S2_group_order"),
                      0)
        return "{}Pass".format(passes)

    block_name_function = {
        "IF": make_if_name, 
        "BBC": make_bbc_name, 
        "PHASE_CAL_DETECT": make_phasecal_name, 
        "FREQ": make_freq_name, 
        "TRACKS": make_tracks_name, 
        "BITSTREAMS": make_bitstreams_name, 
        "DATASTREAMS": make_datastreams_name,
        "ROLL": make_roll_name,
        "PASS_ORDER": make_pass_order_name
    }

    # {block -> {def -> name}}
    block_def_name = {block: {} for block in blocks} 

    # {block -> {def -> {mode -> stations}}}
    block_def_mode_stations = defaultdict(lambda: defaultdict(
        lambda: defaultdict(set)))

    # {scan index -> mode}
    scan_mode_name = {}
    # modes
    modes_text = "$MODE;\n"
    mode_names = set()
    for mode_index, mode in enumerate(modes):
        setup_files = (os.path.split(s)[-1] 
                       for s in mode_index_setups[mode_index])
        stripped = (s[:-4] if s[-4:] == ".set" else s for s in setup_files)
        mode_name = extend_name("+".join(sorted(stripped)), mode_names)
        mode_names.add(mode_name)
        scan_mode_name.update(
            {si: mode_name for si, mi in scan_mode_index.items()
             if mode_index == mi})
        
        modes_text += """
*
def {};
     ref $PROCEDURES = Procedure:{};
""".format(mode_name, ":".join(station.stcode for station in stations))[1:]
        # go through all block in order
        for block in blocks:
            if block not in mode.keys():
                continue
            block_defs = mode[block]
            for block_def, block_stations in block_defs.items():
                block_defs[block_def] = sorted(block_stations)
            # name the defs in order of stations (relevant for name suffix, 
            # when there would otherwise be a name clash)
            for block_def, block_stations in sorted(block_defs.items(),
                                                    key=lambda i: i[1]):
                block_def_mode_stations[block][block_def][mode_name].update(
                    block_stations)
                block_name = block_def_name[block].get(block_def)
                if block_name is None:
                    block_name = extend_name(
                        block_name_function[block](block_def),
                        set(block_def_name[block].values()))
                    block_def_name[block][block_def] = block_name
            # now add the defs to the mode block in block name order
            for block_def, block_stations in \
                sorted(block_defs.items(),
                       key=lambda i: block_def_name[block][i[0]]):
                modes_text += """
     ref ${} = {}{};
""".format(block,  block_def_name[block][block_def], 
           "".join(":" + station for station in block_stations))[1:]
    
        modes_text += "enddef;\n"
    
    # blocks referenced in modes
    for block in blocks:
        block_defs = set()
        for mode in modes:
            block_defs.update(mode[block].keys())
        if len(block_defs) > 0:
            modes_text += block_separator
            modes_text += "${};\n".format(block)
            for block_def in sorted(block_defs, key=lambda block_def: 
                                    block_def_name[block][block_def]):
                modes_text += "*\n{}".format(block_def2str(
                    block_def_name[block][block_def], block_def, 
                    mode_stations=block_def_mode_stations[block][block_def]))
    
    return (modes_text, scan_mode_name)

def procedures_block(vex_version):
    # just a hardcoded procedure
    return """
$PROCEDURES;
*
def Procedure;
     procedure_name_prefix =  "01";
     tape_change =  420 sec;
     headstack_motion =    6 sec;
     new_source_command =    5 sec;
     new_tape_setup =   20 sec;
     setup_always = on :  20 sec;
     parity_check = off : 100 sec;
     tape_prepass = off : 600 sec;
     preob_cal  = on :  10 sec : preob;
     midob_cal  = on :  15 sec : midob;
     postob_cal = on :   0 sec : postob;
enddef;
"""[1:]

def stations_block(vex_version, print_warnings):
    def do_site(station):
        # force 2 char site code
        if len(station.stcode) == 0:
            site_id =  ("site_ID", "Xx")
        elif len(station.stcode) == 1:
            if vex_version >= "2":
                site_id = ("site_ID", station.stcode + station.stcode.lower(),
                           station.stcode)
            else:
                site_id = ("site_ID", station.stcode)
        else:
            site_id = ("site_ID", station.stcode[:2])
        site = (("site_type", "fixed"),
                ("site_name", str2val(station.station)),
                site_id,
                ("site_position",) + tuple(
                    "{:14.5f} m".format(i) for i in 
                    (station.xpos, station.ypos, station.zpos)),
                ("site_velocity",) + tuple(
                    "{:10.6f} m/yr".format(i) for i in
                    (station.dxpos, station.dypos, station.dzpos)),
                ("site_position_epoch", 
                 "{}y{:03d}d".format(*s.timej(station.mjdrate)[:2])))
        if station.nhoriz > 1:
            site += (("horizon_map_az", "{} deg".format(station.horaz[0])) + \
                     tuple(station.horaz[1:]),
                     ("horizon_map_el", "{} deg".format(station.horel[0])) + \
                     tuple(station.horel[1:]))
        return site

    def do_antenna(station):
        mount_map = {"ALTAZ": ("az", "el"),
                     "EQUAT": ("ha", "dec"),
                     "XYNS": ("x", "yns"),
                     "XYEW": ("x", "yew")}
        axes = mount_map.get(station.mount)
        antenna = tuple()
        if axes is not None:
            antenna += (("axis_type",) + axes,
                        ("antenna_motion", axes[0], 
                         "{:6.1f} deg/min".format(station.ax1rate),
                         "{} sec".format(int(station.tsettle))),
                        ("antenna_motion", axes[1], 
                         "{:6.1f} deg/min".format(station.ax2rate),
                         "{} sec".format(int(station.tsettle))))
            if (station.mount == "ALTAZ") and \
               station.station.startswith("VLBA"):
                antenna += pointing_sectors(station, vex_version, 
                                            print_warnings)
            
        antenna += (("axis_offset", "{:10.5} m".format(station.axoff)),)
        return antenna

    def do_das(station):
        das = tuple()
        if station.usedisk:
            if station.disk in ("MARK5" + abc for abc in "ABC"):
                recorder = "Mark5" + station.disk[5]
            elif station.disk == "LBADR":
                recorder = "S2"
            else:
                # replicating SCHED vxwrda.f: use recorder in error output
                # while disk is used for comparison
                raise RuntimeError("Unknown recorder of type: {}".format(
                    station.recorder))
            
            if vex_version < "2":
                das += (("record_transport_type", recorder),)
            else:
                das += (("equip", "recorder", recorder, "&" + recorder),)
                
        dar_map = {"MKIV": "Mark4",
                   "NONE": "none",
                   "R1002": "Mark4",
                   "eMERL": "WIDAR"}
        dar_map.update({k: k for k in ("VLBA", "VLBA4", "K4", "WIDAR", "LBA")})
        if vex_version < "2":
            dar_map.update({"MKIII": "Mark3A",
                            "VLBAG": "VLBAG",
                            "RDBE": "RDBE",
                            "RDBE2": "RDBE2",
                            "DBBC": "DBBC", 
                            "DBBC3": "DBBC"})
        else:
            dar_map.update({"VLBAG": "VLBAG4",
                            "RDBE": "RDBE_DDC",
                            "RDBE2": "RDBE_DDC",
                            "DBBC": "DBBC_DDC",
                            "DBBC3": "DBBC_DDC"})
        try:
            electronics = dar_map[station.dar]
            if vex_version < "2":
                das += (("electronics_rack_type", electronics),)
            else:
                das += (("equip", "rack", electronics, "&" + electronics),)
        except KeyError:
            raise RuntimeError("Unknown DAR of type: {}".format(station.dar))

        if vex_version < "2":
            das += (("number_drives", str(station.stndriv)),)
            if not ((1 <= station.stndriv <= 9) and (1 <= station.nheads <= 9)):
                raise RuntimeError(
                    "Inconsitent number of drives/headstacks")
            for drive in range(station.stndriv):
                for head in range(1, station.nheads + 1):
                    das += (("headstack", head + station.nheads * drive, 
                             "", drive),)

            if station.usetape or (not s.schcon.vextest):
                das += (("tape_motion", "adaptive", 
                         "0 min", "0 min", "10 sec"),)

        return das

    def make_das_name(block_def):
        if vex_version < "2":
            rack = next(param_values[1] for param_values in block_def
                        if param_values[0] == "electronics_rack_type")
            recorder = next((param_values[1] for param_values in block_def
                             if param_values[0] == "record_transport_type"), 
                            None) 
        else:
            rack = next(param_values[2] for param_values in block_def
                        if param_values[1] == "rack")
            recorder = next((param_values[2] for param_values in block_def
                             if param_values[1] == "recorder"), None)
        if recorder is None:
            return rack
        else:
            return "{}+{}".format(rack, recorder)

    stations = StationCatalog().used()
    
    generator = {"SITE": do_site,
                 "ANTENNA": do_antenna,
                 "DAS": do_das}
    block_def_name = {block: {} for block in generator.keys()}
    stations_text = "$STATION;\n"
    for station in stations:
        stations_text += "*\ndef {};\n".format(station.stcode)
        for block, function in sorted(generator.items()):
            block_def = function(station)
            name = block_def_name[block].get(block_def)
            if name is None:
                if block == "DAS":
                    name = extend_name(make_das_name(block_def), 
                                       block_def_name["DAS"].values())
                else:
                    # SITE/ANTENNA, just use station name
                    name = station.station
                name = extend_name(name, block_def_name[block])
                block_def_name[block][block_def] = name
            stations_text += "     ref ${} = {};\n".format(block, name)
        stations_text += "enddef;\n"

    for block, def_name in sorted(block_def_name.items()):
        stations_text += block_separator
        stations_text += "${};\n".format(block)
        for def_, name in sorted(def_name.items(), key=lambda i: i[1]):
            stations_text += "*\n{}".format(block_def2str(name, def_))
            
    return stations_text

def pointing_sectors(station, vex_version, print_warnings):
    ax1 = station.ax1lim
    ax2 = station.ax2lim
    diff = ax1[1] - ax1[0]
    if diff <= 360:
        if ax2[1] <= 90.0001:
            # name, lowaz, highaz, lowel, highel
            zones = (("n",  ax1[0], ax1[1], ax2[0], ax2[1]),)
        else:
            zones = (("n",  ax1[0], ax1[1], ax2[0], 90.),
                     ("np", 0.,     0.,     90.,    ax2[1]))
            if print_warnings:
                s.wlog(1, "WRAPZONE: {} has over-the-top pointing.".format(
                    station.station) )
                s.wlog(1, "          This is not well supported in SCHED.  "
                       "Az and El may be wrong ")
                s.wlog(1, "          and slew calculations may be way off.")
    elif diff <= 720:
        if ax2[1] <= 90.0001:
            zones = (("ccw", ax1[0], ax1[1] - 360, ax2[0], ax2[1]),
                     ("n",   ax1[1] - 360, ax1[0] + 360, ax2[0], ax2[1]),
                     ("cw",  ax1[0] + 360, ax1[0], ax2[0], ax2[1]))
        else:
            zones = (("ccw",  ax1[0], ax1[1] - 360, ax2[0], 90.),
                     ("n",    ax1[1] - 360, ax1[0] + 360, ax2[0], 90.),
                     ("cw",   ax1[0] + 360, ax1[0], ax2[0], 90.),
                     ("ccwp", ax1[0], ax1[1] - 360, 90., ax2[1]),
                     ("ccwp", ax1[1] - 360, ax1[0] + 360, 90., ax2[1]),
                     ("ccwp", ax1[0] + 360, ax1[0], 90., ax2[1]))
    else: # diff > 720
        if print_warnings:
            s.wlog(1, "WRAPZONE: Station {} has an azimuth wrap range over "
                   "180 degrees.".format(station.station))
            s.wlog(1, "          SCHED is not set up for that")
            s.wlog(1, "          Will not give wrap zones in VEX file.")
        
    zone_counter = defaultdict(int)
    def name_suffix(zone):
        zone_counter[zone] += 1
        if zone_counter[zone] == 1:
            return ""
        return "#{}".format(zone_counter[zone])
    deg = "{} deg"
    if vex_version < "2":
        return tuple(("pointing_sector", "&" + zone[0] + name_suffix(zone),
                      "az", deg.format(zone[1]), deg.format(zone[2]),
                      "el", deg.format(zone[3]), deg.format(zone[4])) 
                     for zone in zones)
    else:
        # VEX 2 added a zone name
        return tuple(("pointing_sector", zone[0], 
                      "&" + zone[0] + name_suffix(zone),
                      "az", deg.format(zone[1]), deg.format(zone[2]),
                      "el", deg.format(zone[3]), deg.format(zone[4])) 
                     for zone in zones)

def scan_sector(station, scan, az1, el1):
    if (station.mount == "ALTAZ") and \
       (station.ax1lim[1] - station.ax1lim[0] <= 720):
        if (az1 >= station.ax1lim[0]) and (az1 < station.ax1lim[1] - 360):
            zone = "ccw"
        elif (az1 >= station.ax1lim[0] + 360) and (az1 < station.ax1lim[1]):
            zone = "cw"
        else:
            zone = "n"

        if (station.ax2lim[1] > 90) and (el1 > 90):
            zone += "p"
        return zone
    return None
    
def source_block(vex_version, print_warnings):
    sources = SourceCatalog().used()
    source_text = "$SOURCE;\n"
    source_defs = 0
    for source in sources:
        source_defs += len(source.aliases)
        if (source_defs > 1000) and print_warnings:
            s.wlog(1, "WARNING: More than 1000 sources in this "
                   "schedule. This VEX will NOT run on the Field System!")
        for alias in source.aliases:
            source_text += """
def {};
     source_name = {};
     ra = {}; dec = {}; ref_coord_frame = J2000;
enddef;
""".format(str2def(alias), alias, 
           f2str(s.tformwrp(source.ra2000, "T", 0, 2, 10, "hms")),
           f2str(s.tformwrp(source.d2000,  " ", 1, 2, 9,  "d'\"")))[1:]
            
    return source_text

def sched_block(scan_mode, vex_version, print_warnings):
    catalog = ScanCatalog()
    scans = catalog.used()
    scan_offset = catalog.scan_offset
    
    stations = StationCatalog().used()

    setups = SetupCatalog().scheduled()

    byte_offset = defaultdict(float) # station -> expected bytes recorded

    ret = "$SCHED;\n"
    for scan_index, scan in enumerate(scans):
        found_one = False
        for station in stations:
            setup = setups[station.nsetup[scan_index+scan_offset] - 1]
            if station.stascn[scan_index+scan_offset] and \
               (station.station.startswith("VLBA") or setup.format != "NONE"):
                found_one = True
                break
        if not found_one:
            continue

        scan_def = (("start", time2str(scan.startj)),
                    ("mode", scan_mode[scan_index]),
                    ("source", scan.scnsrc))
        if (vex_version >= "2") and (len(scan.scanexps) > 0):
            scan_def += (
                ("intent", "", "EXPERIMENTS", ",".join(scan.scanexps)),)
        scan_name = "No{:04d}".format(scan_index+scan_offset+1)
        for station in stations:
            if station.stascn[scan_index+scan_offset]:
                setup = setups[station.nsetup[scan_index+scan_offset] - 1]
                if (scan.grabto == "FILE") and (scan.datapath == "IN2NET"):
                    raise RuntimeError(
                        "You have requested a GRABTO (ftp) scan, but you are "
                        "not recording to disk. You must set DATAPATH=IN2DISK")
                
                if scan.grabto == "NET":
                    if print_warnings:
                        s.wlog(1, "You have requested GRABTO=NET, but that "
                               "is not supported in VEX and will be ignored. ")

                elif scan.grabto == "FILE":
                    grab_duration, before_stop = scan.grabtime
                    scan_duration = (scan.stopj - scan.startj) * \
                                    parameter.secpday
                    if grab_duration + before_stop > scan_duration:
                        raise RuntimeError(
                            "You have scheduled a GRABTO scan, but the "
                            "GRABTIME is not consistent with the scan length. "
                            "Please increase the scan length or change the "
                            "GRABTIME parameters.")
                    # cast from numpy.float to python float so round returns int
                    grab_stop = round(float(scan_duration-before_stop))
                    grab_start = round(float(grab_stop-grab_duration))
                    scan_def += (("data_transfer", 
                                  station.stcode,
                                  "disk2file",
                                  "{}_{}_{}".format(
                                      f2str(s.schc1.expcode).lower(),
                                      station.stcode.lower(),
                                      scan_name.lower()),
                                  "{} sec".format(grab_start),
                                  "{} sec".format(grab_stop)),)
                    if scan_index < len(scans) - 1:
                        # check that enough time to do the disk2file is 
                        # schedule before the next scan
                        bitrate = setup.totbps
                        # assume a 110Mbps disk2file rate and some latency
                        expected_time = 5 + grab_duration * bitrate / 110
                        gap = (scans[scan_index+1].startj - scan.stopj) \
                              * parameter.secpday
                        if (expected_time > gap) and print_warnings:
                            s.wlog(1, "You have scheduled an ftp "
                                   "(GRABTO) scan but you have not left a long "
                                   "enough gap to transfer the data before the "
                                   "next scan starts. Try inserting a gap of "
                                   "{} seconds after the ftp scan (#{})".format(
                                       expected_time + 1, scan_index + 1))
                elif scan.datapath == "IN2NET":
                    # cast from numpy.float to python float so round returns int
                    duration = round(float(scan.stopj - scan.startj) * 
                                     parameter.secpday)
                    scan_def += (("data_transfer",
                                  station.stcode,
                                  "in2net",
                                  "",
                                  "0 sec",
                                  "{} sec".format(duration)),)

                if station.station.startswith("VLBA") or \
                   (setup.format != "NONE"):
                    # cast from numpy.float to python float so round returns int
                    data_good = round(max(
                        float(station.tonsrc[scan_index+scan_offset] - 
                              scan.startj) * parameter.secpday,
                        0))
                    data_stop = round(float(scan.stopj - scan.startj) * 
                                      parameter.secpday)
                    if station.usedisk:
                        media_position = "{:9.3f} GB".format(
                            byte_offset[station.stcode])
                        byte_offset[station.stcode] = station.gbytes[
                            scan_index+scan_offset]
                    else:
                        media_position = ""

                    pointing_sector = ""
                    if station.station.startswith("VLBA"):
                        zone = scan_sector(
                            station, scan,
                            station.az1[scan_index+scan_offset],
                            station.el1[scan_index+scan_offset])
                        # assumption here is that the zone name is the link name
                        # according to comments in sched this is true in the
                        # normal case, otherwise 3 zones have the same name,
                        # no clue which one would be the one to pick
                        # see Sched/wrapzone.f
                        if zone is not None:
                            pointing_sector = "&" + zone
                        
                    pass_ = ""
                    if vex_version < "2" and station.disk == "LBADR":
                        pass_ = 0
                    scan_def += (("station",
                                  station.stcode,
                                  "{} sec".format(min(data_good, data_stop)),
                                  "{} sec".format(data_stop),
                                  media_position,
                                  pass_,
                                  pointing_sector,
                                  0 if scan.norec else 1),)

        ret += "*\n{}".format(
            block_def2str(scan_name, scan_def, keyword="scan"))

    if print_warnings:
        frequency_setups = FrequencyCatalog().scheduled()
        vex_scans.check_tsys(scans, scan_offset, stations, setups, 
                             frequency_setups)

        warn_field_system = vex_scans.check_minimum_scan_duration(
            scans, scan_offset, stations) \
            or vex_scans.check_scan_overlap(scans, scan_offset, stations) \
            or vex_scans.check_mode_change(scans, scan_offset, stations,
                                                scan_mode)
            
        if warn_field_system:
            s.wlog(1, "WARNING: Scan timing problem for PCFS, "
                   "this VEX will NOT run!!!!")

        vex_scans.check_recording_sizes(scans, scan_offset, stations)
    
    return ret
    
def block_def2str(name, block_def, keyword="def", mode_stations={}):
    ret = """
{} {};
""".format(keyword, name)[1:]

    # add a comment in which mode this is referenced by which stations
    for mode, stations in sorted(mode_stations.items()):
        ret += """
* {}: {}
""".format(mode, ", ".join(sorted(stations)))[1:]
    
    # align columns for the same param value lines
    param_columns = {param_value[0]: defaultdict(int) 
                     for param_value in block_def}
    for param_value in block_def:
        param = param_value[0]
        for column, value in enumerate(param_value[1:]):
            param_columns[param][column] = \
                max(param_columns[param][column], len(str(value)))
    for param_value in block_def:
        param = param_value[0]
        ret += "     {} = {};\n".format(
            param,
            " : ".join("{:>{width}}".format(value, 
                                            width=param_columns[param][column])
                       for column, value in enumerate(param_value[1:])))
    ret += "end{};\n".format(keyword)
    return ret

def str2def(s):
    s = s.strip()
    s = re.sub("\s", "_", s)
    s = re.sub("[;:=]", "X", s)
    return s[:32]

def str2val(s):
    if (len(s) == 0) or re.match(".*[\s;:=&*$].*", s):
        # pre-undo all standard C-quoting rules
        s = s.replace('\\', '\\\\')
        s = s.replace('"', '\\"')
        return '"{}"'.format(s)
    return s

def text2comments(lines, comment_prefix):
    # comments may only contain 128 characters, including leading *
    # try to split text on word boundary when required
    ret = ""
    max_comment_length = 128 - len(comment_prefix)
    for line in lines:
        if line:
            if len(line) > max_comment_length:
                index = line[: max_comment_length].rfind(" ")
                if index <  len(line) - max_comment_length:
                    # split in the middle of a word
                    index = max_comment_length
                split_lines = [line[:index], line[index:]]
            else:
                split_lines = [line]
            for split_line in split_lines:
                ret += comment_prefix + split_line + "\n"
    return ret

def time2str(t):
    year, doy, time_ = s.timej(t)
    return "{:}y{:03d}d{}".format(year, doy, f2str(
        s.tformwrp(time_, "T", 0, 2, 2, "hms")))

def extend_name(base_name, used_names):
    extension = 2
    # def is at most 32 chars, reserve 3 for the extension
    base_name = str2def(base_name)[:29]
    
    name = base_name
    while name in used_names:
        name = base_name + "#{:02d}".format(extension)
        extension += 1
    return name




from . import scndup
from ..catalog import ScanCatalog

import schedlib as s

scan_catalog = ScanCatalog()
def optnone(k_scan, scan_index):
    """
    Copies scan indexed by k_scan to scan indexed by scan_index,
    if the indices differ and not done already yet.
    Return tuple of booleans (adjust, keep, done)
    """
    if s.schcon.debug and (k_scan <= 3):
        s.wlog(0, "OPTNONE: Starting.")

    done = (k_scan > s.schn1.nscans)
    if (not done) and (k_scan != scan_index):
        scndup(scan_index - 1, k_scan - 1, True, "OPTNONE")

    # duronly, 1 means only start times, 4 only stop times specified
    adjust = (s.schn1.dwells and (scan_index > s.schn1.scan1)) and \
             (scan_catalog.direct_access_entries[scan_index - 1].duronly in 
              {1, 4})
    return (adjust, True, done)
        



from . import scndup
from .parameter import secpday
from ..catalog import StationCatalog, ScanCatalog
from ..util import f2str

import schedlib as s

import numpy as np

def gmkscn(last_scan_index, scan_index, j_scan, source_index, source_name,
           approx_time, min_elevation, keep_station, scan_stascn, g_mode):
    if s.schcon.debug:
        s.wlog(0, "GMKSCN starting.")

    scan_catalog = ScanCatalog()
    station_catalog = StationCatalog()
    
    scans = scan_catalog.direct_access_entries
    stations = station_catalog.used(use_direct_access=True)

    n_good = 0

    # go to python/zero indexing for scan index
    scan_index -= 1
    j_scan -= 1

    scndup(scan_index, j_scan, False, "GMKSCN")

    scan = scans[scan_index]

    scan.srcnum = source_index
    scan.scnsrc = source_name

    ok_sta = np.empty(dtype=bool, shape=(len(stations),))
    if g_mode == "SET":
        for i, station in enumerate(stations):
            ok_sta[i] = False
            if station.stascn[j_scan]:
                last_time, available_time = s.stageo(
                    scan_index + 1, i + 1, approx_time, last_scan_index[i], 
                    "GMKSCN")
                if (f2str(station.up1[scan_index]) == "") and \
                   (f2str(station.up2[scan_index]) == "") and \
                   (station.el1[scan_index] > min_elevation) and \
                   (station.el2[scan_index] > min_elevation):
                    ok_sta[i] = True
                    n_good += 1

        for i, station in enumerate(stations):
            station.stascn[scan_index] = station.stascn[j_scan] and ok_sta[i]

        if n_good >= 1:
            scan.startj = approx_time
            s.opttim(last_scan_index, last_scan_index, scan_index + 1, 
                     True, False, False)
            n_good = s.scngeo(last_scan_index, scan_index + 1)

        if (n_good >= scans[j_scan].opmian) and (scan_index + 1 > s.schn1.scan1):
            in_scan = [s for s in stations if s.stascn[scan_index]]
            in_scan.sort(key=lambda s: -s.tonsrc[scan_index])
            if len(in_scan) > 0:
                t_med = in_scan[min(2, len(in_scan) - 1)].tonsrc[scan_index]
            else:
                s.errlog("GMKSCN: INSCN zero.  Programming error")

            for i, station in enumerate(stations):
                if station.stascn[scan_index] and \
                   (i + 1 != keep_station) and \
                   (last_scan_index[i] != 0) and \
                   (station.tonsrc[scan_index] > 
                    t_med + s.schsou.geoslow / secpday):
                    if station.el1[scan_index] < s.schsou.geolowel:
                        n_good = 0
                        if s.schsou.geoprt >= 2:
                            s.wlog(0, "**gmkscn: Dropping scan - low el for "
                                   "slow ant {} {}  {} {}".format(
                                       i + 1, scan_index + 1, source_name, 
                                       station.el1[scan_index]))
                    else:
                        n_good = max(n_good - 1, 0)
                        station.stascn[scan_index] = False
                        ok_sta[i] = False
                        if s.schsou.geoprt >= 2:
                            s.wlog(0, "++gmkscn: Dropping station {}  Scan {} {}"
                                   "  Geosrc: {} for long slew.".format(
                                       i + 1, scan_index + 1, s.schn1.scan1, 
                                       source_name))

        scan_stascn[:] = [s.stascn[scan_index] for s in stations]

    elif g_mode == "FORCE":
        for i, station in enumerate(stations):
            station.stascn[scan_index] = scan_stascn[i]
        ok_sta[:] = scan_stascn[:ok_sta.shape[0]]
        n_good = np.count_nonzero(ok_sta)

        if n_good >= 1:
            scan.startj = approx_time
            s.opttim(last_scan_index, last_scan_index, scan_index + 1, 
                     True, False, False)
            n_good = s.scngeo(last_scan_index, scan_index + 1)
    else:
        s.errlog("GMKSCN: Bad GMODE.  Programming error.")

    return n_good, ok_sta, scan_stascn
          



# there is a potential problem with cyclic imports here 
# (eg. input calls schfiles), but so far I haven't found a fortran call cycle,
# so with the right order, exposing the module functions like this works fine
from .schdefs_module  import schdefs
from .stread_module   import stread
from .srread_module   import srread, srread_open
from .getfreq_module  import getfreq
from .getsta_module   import getsta
from .rfreq_module    import rfreq
from .rdpeak_module   import rdpeak, rdpeak_open
from .satini_module   import satini
from .pcread_module   import pcread
from .rdset_module    import rdset
from .schfiles_module import schfiles
from .gettim_module   import gettim
from .gintent_module  import gintent
from .toggle_module   import toggle
from .infdb_module    import infdb
from .invla_module    import invla
from .scndup_module   import scndup
from .schrep_module   import schrep
from .getcov_module   import getcov
from .getcor_module   import getcor
from .times_module    import times
from .sttant_module   import sttant
from .schin_module    import schin
from .getset_module   import getset
from .input_module    import input_
from .bbcalt_module   import bbcalt
from .ifdbbc_module   import ifdbbc
from .ifdbbc3_module  import ifdbbc3
from .bbcdbbc_module  import bbcdbbc
from .bbcemerlin_module import bbc_emerlin
from .setbbc_module   import setbbc
from .setdefs_module  import setdefs
from .chkdbbc_module  import chkdbbc
from .checkemerlin_module import check_emerlin
from .chkset_module   import chkset
from .setexpnd_module import setexpnd
from .defset_module   import defset
from .defaults_module import defaults
from .vexout_module   import vexout
from .gmkscn_module   import gmkscn
from .makeseg_module  import makeseg
from .optnone_module  import optnone
from .optskd_module   import optskd
from .optcells_module import optcells
from .optupt_module   import optupt
from .opthas_module   import opthas
from .opthiel_module  import opthiel
from .addpeak_module  import addpeak
from .makescn_module  import makescn
from .geochk_module   import geochk
from .geomake_module  import geomake
from .addgeo_module   import addgeo
from .schopt_module   import schopt
from .stmsg_module    import stmsg
